"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

###########################################################################
# NOTE: Early Adopter implemented based on 1.226. Check back for updates. #
###########################################################################

from typing import List, Dict, Any, Optional, Union
from enum import Enum
from datetime import datetime

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList
from dynatrace.utils import int64_to_datetime, datetime_to_int64, timestamp_to_string
from dynatrace.environment_v2.custom_tags import METag
from dynatrace.environment_v2.monitored_entities import EntityStub
from dynatrace.environment_v2.schemas import ManagementZone


class EventServiceV2:
    ENDPOINT_EVENTS = "/api/v2/events"
    ENDPOINT_INGEST = "/api/v2/events/ingest"
    ENDPOINT_TYPES = "/api/v2/eventTypes"

    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(
        self,
        page_size: Optional[int] = None,
        time_from: Optional[Union[datetime, str]] = None,
        time_to: Optional[Union[datetime, str]] = None,
        event_selector: Optional[str] = None,
        entity_selector: Optional[str] = None,
    ) -> "PaginatedList[Event]":
        """Lists events within the specified timeframe

        :param page_size: The amount of events in a single response payload. The maximal allowed page size is 1000. If not set, 100 is used.
        :param time_from: The start of the requested timeframe. If not set, the relative timeframe of two hours is used (now-2h).
            You can use one of the following formats:
            * Timestamp in UTC milliseconds.
            * Human-readable format of 2021-01-25T05:57:01.123+01:00. If no time zone is specified, UTC is used.
            * Relative timeframe, back from now. The format is now-NU/A, where N is the amount of time, U is the unit of time, and A is an alignment.

        :param time_to: The end of the requested timeframe. If not set, the current timestamp is used. Same formats as above.
        :param event_selector: Defines the event scope of the query.
        :param entity_selector: The entity scope of the query. The length of the string is limited to 10,000 characters.
            The number of entities that can be selected is limited to 10,000. For more information see https://dt-url.net/apientityselector

        :returns PaginatedList[Event]: the list of events
        """
        params = {
            "pageSize": page_size,
            "from": timestamp_to_string(time_from),
            "to": timestamp_to_string(time_to),
            "eventSelector": event_selector,
            "entitySelector": entity_selector,
        }
        return PaginatedList(target_class=Event, http_client=self.__http_client, target_url=self.ENDPOINT_EVENTS, list_item="events", target_params=params)

    def get(self, event_id: str) -> "Event":
        """Gets the properties of an event referenced by ID.

        :param event_id: The ID of the required event.

        :returns Event: the requested event
        """
        response = self.__http_client.make_request(path=f"{self.ENDPOINT_EVENTS}/{event_id}")
        return Event(raw_element=response.json(), http_client=self.__http_client)

    def list_types(self, page_size: Optional[int] = None) -> "PaginatedList[EventType]":
        """Lists all event types.

        :param page_size: The amount of event types in a single response payload. The maximal allowed page size is 500. If not set, 100 is used.

        :returns PaginatedList[EventType]: the list of event types
        """
        params = {"pageSize": page_size}
        return PaginatedList(
            target_class=EventType, http_client=self.__http_client, target_url=self.ENDPOINT_TYPES, list_item="eventTypeInfos", target_params=params
        )

    def get_type(self, event_type: str) -> "EventType":
        """Gets the properties of a specific event type.

        :param event_type: The event type you're inquiring.

        :returns EventType: the event type requested
        """
        response = self.__http_client.make_request(path=f"{self.ENDPOINT_TYPES}/{event_type}")
        return EventType(raw_element=response.json(), http_client=self.__http_client)

    def ingest(
        self,
        event_type: str,
        title: str,
        start_time: Optional[Union[datetime, str]] = None,
        end_time: Optional[Union[datetime, str]] = None,
        timeout: int = 15,
        entity_selector: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None
    ):
        """
        Ingests a custom event to Dynatrace

        :param event_type: the type of event
        :param title: the title of the event
        :param start_time: the start time of the event
        :param end_time: the end time of the event
        :param timeout: the timeout of the event in minutes
        :param entity_selector: the entity selector, defining a set of Dynatrace entities to be associated with the event
        :param properties: a map of event properties
        """
        
        params = {
            "eventType": event_type,
            "title": title,
            "startTime": datetime_to_int64(start_time),
            "endTime": datetime_to_int64(end_time),
            "timeout": timeout,
            "entitySelector": entity_selector,
            "properties": properties
        }

        return self.__http_client.make_request(f"{self.ENDPOINT_INGEST}", method="POST", params=params).json()


class Event(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        # Mandatory
        self.event_id: str = raw_element["eventId"]
        self.event_type: str = raw_element["eventType"]
        self.start_time: datetime = int64_to_datetime(int(raw_element["startTime"]))
        # Optional
        self.entity_tags: Optional[List[METag]] = [METag(raw_element=et) for et in raw_element.get("entityTags", [])]
        self.suppress_alert: Optional[bool] = raw_element.get("suppressAlert")
        self.frequent_event: Optional[bool] = raw_element.get("frequentEvent")
        self.suppress_problem: Optional[bool] = raw_element.get("suppressProblem")
        self.under_maintenance: Optional[bool] = raw_element.get("underMaintenance")
        self.entity_id: Optional[EntityStub] = EntityStub(raw_element=raw_element["entityId"]) if raw_element.get("entityId") else None
        self.management_zones: Optional[List[ManagementZone]] = [ManagementZone(raw_element=mz) for mz in raw_element.get("managementZones", [])]
        self.properties: Optional[List[EventProperty]] = [EventProperty(raw_element=p) for p in raw_element.get("properties", [])]
        self.status: Optional[EventStatus] = EventStatus(raw_element.get("status")) if raw_element.get("status") else None
        self.title: Optional[str] = raw_element.get("title")
        self.correlation_id: Optional[str] = raw_element.get("correlationId")
        self.end_time: Optional[datetime] = int64_to_datetime(raw_element["endTime"]) if raw_element["endTime"] != -1 else None

    def to_json(self) -> Dict[str, Any]:
        return {
            "eventId": self.event_id,
            "eventType": self.event_type,
            "startTime": datetime_to_int64(self.start_time),
            "endTime": datetime_to_int64(self.end_time),
            "entityTags": [tag.to_json() for tag in self.entity_tags] if self.entity_tags else [],
            "suppressAlert": self.suppress_alert,
            "frequentEvent": self.frequent_event,
            "suppressProblem": self.suppress_problem,
            "underMaintenance": self.under_maintenance,
            "entityId": self.entity_id.to_json() if self.entity_id else None,
            "managementZones": [mz.to_json() for mz in self.management_zones] if self.management_zones else [],
            "properties": [p.to_json() for p in self.properties] if self.properties else [],
            "title": self.title,
            "correlationId": self.correlation_id,
        }


class EventProperty(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.key: Optional[str] = raw_element.get("key")
        self.value: Optional[str] = raw_element.get("value")

    def to_json(self) -> Dict[str, Any]:
        return {"key": self.key, "value": self.value}


class EventType(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.type: str = raw_element["type"]
        self.display_name: str = raw_element["displayName"]
        self.severity_level: Optional[EventSeverity] = EventSeverity(raw_element["severityLevel"]) if raw_element.get("severityLevel") else None
        self.description: Optional[str] = raw_element.get("description")

    def to_json(self) -> Dict[str, Any]:
        return {
            "type": self.type,
            "displayName": self.display_name,
            "severityLevel": str(self.severity_level) if self.severity_level else None,
            "description": self.description,
        }


class EventStatus(Enum):
    OPEN = "OPEN"
    CLOSED = "CLOSED"

    def __str__(self) -> str:
        return self.value


class EventSeverity(Enum):
    AVAILABILITY = "AVAILABILITY"
    CUSTOM_ALERT = "CUSTOM_ALERT"
    ERROR = "ERROR"
    INFO = "INFO"
    MONITORING_UNAVAILABLE = "MONITORING_UNAVAILABLE"
    PERFORMANCE = "PERFORMANCE"
    RESOURCE_CONTENTION = "RESOURCE_CONTENTION"

    def __str__(self) -> str:
        return self.value



from typing import Optional, Dict, Any, List, Union
from datetime import datetime

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList
from dynatrace.utils import int64_to_datetime


class SettingService:
    OBJECTS_ENDPOINT = "/api/v2/settings/objects"
    SCHEMAS_ENDPOINT = "/api/v2/settings/schemas"

    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client
        

    def list_schemas(self) -> PaginatedList["SchemaStub"]:
        """Lists all settings schemas available in your environment"""

        return PaginatedList(
            SchemaStub,
            self.__http_client,
            target_url=self.SCHEMAS_ENDPOINT,
            list_item="items"
        )


    def list_objects(
        self,
        schema_id: Optional[str] = None,
        scope: Optional[str] = None,
        external_ids: Optional[str] = None,
        fields: Optional[str] = None,
        filter: Optional[str] = None,
        sort: Optional[str] = None,
        page_size: Optional[str] = None,
    ) -> PaginatedList["SettingsObject"]:
        """Lists settings

        :return: a list of settings with details
        """
        params = {
            "schemaIds": schema_id,
            "scope": scope,
            "fields": fields,
            "externalIds": external_ids,
            "filter": filter,
            "sort": sort,
            "pageSize": page_size,
        }
        return PaginatedList(
            SettingsObject,
            self.__http_client,
            target_url=self.OBJECTS_ENDPOINT,
            list_item="items",
            target_params=params,
        )

    def create_object(
        self,
        validate_only: Optional[bool] = False,
        body: Union[
            Optional[List["SettingsObjectCreate"]], Optional["SettingsObjectCreate"]
        ] = [],
    ):
        """
        Creates a new settings object or validates the provided settigns object

        :param validate_only: If true, the request runs only validation of the submitted settings objects, without saving them
        :param body: The JSON body of the request. Contains the settings objects
        """
        query_params = {"validateOnly": validate_only}

        if isinstance(body, SettingsObjectCreate):
            body = [body]

        body = [o.json() for o in body]

        response = self.__http_client.make_request(
            self.OBJECTS_ENDPOINT, params=body, method="POST", query_params=query_params
        ).json()
        return response

    def get_object(self, object_id: str):
        """Gets parameters of specified settings object

        :param object_id: the ID of the object
        :return: a Settings object
        """
        response = self.__http_client.make_request(
            f"{self.OBJECTS_ENDPOINT}/{object_id}"
        ).json()
        return SettingsObject(raw_element=response)

    def update_object(
        self, object_id: str, body: Optional["SettingsObjectUpdate"] = None
    ):
        """Updates an existing settings object

        :param object_id: the ID of the object
        :param value: the JSON body of the request. Contains updated parameters of the settings object.
        """
        return self.__http_client.make_request(
            f"{self.OBJECTS_ENDPOINT}/{object_id}", params=body.json(), method="PUT"
        )

    def delete_object(self, object_id: str, update_token: Optional[str] = None):
        """Deletes the specified object

        :param object_id: the ID of the object
        :param update_token: The update token of the object. You can use it to detect simultaneous modifications by different users
        :return: HTTP response
        """
        query_params = {"updateToken": update_token}
        return self.__http_client.make_request(
            f"{self.OBJECTS_ENDPOINT}/{object_id}",
            method="DELETE",
            query_params=query_params,
        )


class ModificationInfo(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.deleteable: bool = raw_element.get("deleteable")
        self.first: bool = raw_element.get("first")
        self.modifiable: bool = raw_element.get("modifiable")
        self.modifiable_paths: List[str] = raw_element.get("modifiablePaths", [])
        self.movable: bool = raw_element.get("movable")
        self.non_modifiable_paths: List[str] = raw_element.get("nonModifiablePaths", [])


class SettingsObject(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        # Mandatory
        self.object_id: str = raw_element["objectId"]
        self.value: dict = raw_element["value"]
        # Optional
        self.author: str = raw_element.get("author")
        self.created: datetime = (
            int64_to_datetime(int(raw_element.get("created")))
            if raw_element.get("created")
            else None
        )
        self.created_by: str = raw_element.get("createdBy")
        self.external_id: str = raw_element.get("externalId")
        self.modification_info: ModificationInfo = (
            ModificationInfo(
                self._http_client, self._headers, raw_element.get("modificationInfo")
            )
            if raw_element.get("modificationInfo")
            else None
        )
        self.modified: datetime = (
            int64_to_datetime(int(raw_element.get("modified")))
            if raw_element.get("modified")
            else None
        )
        self.modified_by: str = raw_element.get("modifiedBy")
        self.schema_id: str = raw_element.get("schemaId")
        self.schema_version: str = raw_element.get("schemaVersion")
        self.scope: str = raw_element.get("scope")
        self.search_summary: str = raw_element.get("searchSummary")
        self.summary: str = raw_element.get("summary")
        self.update_token: str = raw_element.get("updateToken")


class SettingsObjectCreate:
    def __init__(
        self,
        schema_id: str,
        value: dict,
        scope: str,
        external_id: Optional[str] = None,
        insert_after: Optional[str] = None,
        object_id: Optional[str] = None,
        schema_version: Optional[str] = None,
    ):
        self.schema_id = schema_id
        self.value = value
        self.scope = scope
        self.external_id = external_id
        self.insert_after = insert_after
        self.object_id = object_id
        self.schema_version = schema_version

    def json(self) -> dict:
        body = {"schemaId": self.schema_id, "value": self.value, "scope": self.scope}
        if self.external_id:
            body["externalId"] = self.external_id
        if self.insert_after:
            body["insertAfter"] = self.insert_after
        if self.object_id:
            body["objectId"] = self.object_id
        if self.schema_version:
            body["schemaVersion"] = self.schema_version
        return body


class SettingsObjectUpdate:
    def __init__(
        self,
        value: dict,
        insert_after: Optional[str] = None,
        insert_before: Optional[str] = None,
        schema_version: Optional[str] = None,
        update_token: Optional[str] = None,
    ):
        self.value = value
        self.insert_after = insert_after
        self.insert_before = insert_before
        self.schema_version = schema_version
        self.update_token = update_token

    def json(self) -> dict:
        body = {"value": self.value}
        if self.insert_after:
            body["insertAfter"] = self.insert_after
        if self.insert_before:
            body["insertBefore"] = self.insert_before
        if self.schema_version:
            body["schemaVersion"] = self.schema_version
        if self.update_token:
            body["updateToken"] = self.update_token
        return body


class SchemaStub(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.display_name = raw_element["displayName"]
        self.latest_schema_version = raw_element["latestSchemaVersion"]
        self.schema_id = raw_element["schemaId"]



from dynatrace import Dynatrace
from datetime import datetime

from dynatrace.environment_v2.logs import LogRecord, EventType, LogRecordStatus
from dynatrace.pagination import PaginatedList


def test_export(dt: Dynatrace):
    logs = dt.logs.export(time_from="now-10m")
    assert isinstance(logs, PaginatedList)

    logs = list(logs)
    assert len(logs) == 18

    first = logs[0]
    assert isinstance(first, LogRecord)
    assert first.additional_columns['dt.extension.ds'][0] == "python"
    assert first.content.startswith("Failed to assign")
    assert first.event_type == EventType.SFM
    assert first.status == LogRecordStatus.ERROR
    assert first.timestamp == datetime.utcfromtimestamp(1683574915193 / 1000)




"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pprint
from typing import Any, Dict, Optional

from requests import Response

from dynatrace.http_client import HttpClient


class DynatraceObject:
    def __init__(self,
                 http_client: Optional[HttpClient] = None,
                 headers: Optional[Dict[str, str]] = None,
                 raw_element: Optional[Dict[str, Any]] = None):
        if raw_element is None:
            raw_element = {}
        self._http_client = http_client
        self._headers = headers
        self._raw_element = raw_element
        self._create_from_raw_data(raw_element)

    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        pass

    def __repr__(self):
        return f"{self.__class__.__name__}({pprint.pformat(self._raw_element, width=130)})"

    def _make_request(self,
                      path: str,
                      params: Optional[Dict] = None,
                      headers: Optional[Dict] = None,
                      method = "GET",
                      data = None) -> Response:
        return self._http_client.make_request(path, params, headers, method, data)

    def json(self):
        return self._raw_element



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Dict, Any
from enum import Enum

from dynatrace.environment_v2.schemas import ConfigurationMetadata
from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.environment_v2.monitored_entities import EntityShortRepresentation
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList
from dynatrace.configuration_v1.auto_tags import (
    ComparisonBasicType,
    ConditionKeyAttribute,
    ConditionKeyType,
    PropagationType,
)


class RuleType(Enum):
    APPMON_SERVER = "APPMON_SERVER"
    APPMON_SYSTEM_PROFILE = "APPMON_SYSTEM_PROFILE"
    AWS_ACCOUNT = "AWS_ACCOUNT"
    AWS_APPLICATION_LOAD_BALANCER = "AWS_APPLICATION_LOAD_BALANCER"
    AWS_AUTO_SCALING_GROUP = "AWS_AUTO_SCALING_GROUP"
    AWS_CLASSIC_LOAD_BALANCER = "AWS_CLASSIC_LOAD_BALANCER"
    AWS_NETWORK_LOAD_BALANCER = "AWS_NETWORK_LOAD_BALANCER"
    AWS_RELATIONAL_DATABASE_SERVICE = "AWS_RELATIONAL_DATABASE_SERVICE"
    AZURE = "AZURE"
    BROWSER_MONITOR = "BROWSER_MONITOR"
    CLOUD_APPLICATION = "CLOUD_APPLICATION"
    CLOUD_APPLICATION_NAMESPACE = "CLOUD_APPLICATION_NAMESPACE"
    CLOUD_FOUNDRY_FOUNDATION = "CLOUD_FOUNDRY_FOUNDATION"
    CUSTOM_APPLICATION = "CUSTOM_APPLICATION"
    CUSTOM_DEVICE = "CUSTOM_DEVICE"
    CUSTOM_DEVICE_GROUP = "CUSTOM_DEVICE_GROUP"
    DATA_CENTER_SERVICE = "DATA_CENTER_SERVICE"
    ENTERPRISE_APPLICATION = "ENTERPRISE_APPLICATION"
    ESXI_HOST = "ESXI_HOST"
    EXTERNAL_MONITOR = "EXTERNAL_MONITOR"
    HOST = "HOST"
    HOST_GROUP = "HOST_GROUP"
    HTTP_MONITOR = "HTTP_MONITOR"
    KUBERNETES_CLUSTER = "KUBERNETES_CLUSTER"
    KUBERNETES_SERVICE = "KUBERNETES_SERVICE"
    MOBILE_APPLICATION = "MOBILE_APPLICATION"
    MULTIPROTOCOL_MONITOR = "MULTIPROTOCOL_MONITOR"
    OPENSTACK_ACCOUNT = "OPENSTACK_ACCOUNT"
    PROCESS_GROUP = "PROCESS_GROUP"
    QUEUE = "QUEUE"
    SERVICE = "SERVICE"
    WEB_APPLICATION = "WEB_APPLICATION"
    NONE = None


class ManagementZoneService:
    ENDPOINT = "/api/config/v1/managementZones"

    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(
        self, page_size: int = 200
    ) -> PaginatedList["ManagementZoneShortRepresentation"]:
        """
        List all management zones.

        :param page_size: The number of results per result page. Must be between 1 and 500
            Default value : 200
        """
        params = {"pageSize": page_size}
        return PaginatedList(
            ManagementZoneShortRepresentation,
            self.__http_client,
            f"{self.ENDPOINT}",
            params,
            list_item="values",
        )

    def get(self, management_zone_id: str) -> "ManagementZone":
        """Gets the description of a management zone referenced by ID.

        :param _id: The ID of the required management zone.

        :returns Event: the requested management zone
        """
        response = self.__http_client.make_request(
            path=f"{self.ENDPOINT}/{management_zone_id}"
        )
        return ManagementZone(
            raw_element=response.json(), http_client=self.__http_client
        )

    def delete(self, management_zone_id: str):
        """Deletes the specified management zone

        :param networkzone_id: the ID of the management zone
        :return: HTTP response
        """
        return self.__http_client.make_request(
            path=f"{self.ENDPOINT}/{management_zone_id}", method="DELETE"
        )


class ComparisonBasic(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.operator: str = raw_element.get("operator")
        self.value: dict = raw_element.get("value")
        self.negate: bool = raw_element.get("negate")
        self.type: ComparisonBasicType = ComparisonBasicType(raw_element.get("type"))
        self.case_sensitive: bool = raw_element.get("caseSensitive")


class ConditionKey(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.attribute: ConditionKeyAttribute = ConditionKeyAttribute(
            raw_element.get("attribute")
        )
        self.type: ConditionKeyType = ConditionKeyType(raw_element.get("type"))


class EntityRuleEngineCondition(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.key: ConditionKey = ConditionKey(raw_element=raw_element.get("key"))
        self.comparison_info: ComparisonBasic = ComparisonBasic(
            raw_element=raw_element.get("comparisonInfo")
        )


class EntitySelectorBasedManagementZoneRule(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.enabled: bool = raw_element.get("enabled")
        self.entity_selector: str = raw_element.get("entitySelector")
        self.value_format: str = raw_element.get("valueFormat")


class ManagementZoneRule(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.type: RuleType = RuleType(raw_element.get("type"))
        self.enabled: bool = raw_element.get("enabled")
        self.value_format: str = raw_element.get("valueFormat")
        self.propagation_types: List[PropagationType] = [
            PropagationType(prop_type)
            for prop_type in (raw_element.get("propagationTypes") or [])
        ]
        self.conditions: List[EntityRuleEngineCondition] = [
            EntityRuleEngineCondition(raw_element=condition)
            for condition in (raw_element.get("conditions") or [])
        ]


class ManagementZone(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.metadata: ConfigurationMetadata = ConfigurationMetadata(
            self._http_client, None, raw_element.get("metadata")
        )
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.description: str = raw_element.get("description")
        self.rules: List[ManagementZoneRule] = [
            ManagementZoneRule(raw_element=rule) for rule in raw_element.get("rules")
        ]
        self.entity_selector_based_rules: List[
            EntitySelectorBasedManagementZoneRule
        ] = [
            EntitySelectorBasedManagementZoneRule(raw_element=rule)
            for rule in (raw_element.get("entitySelectorBasedRules") or [])
        ]


class ManagementZoneShortRepresentation(EntityShortRepresentation):
    def get_full_configuration(self):
        """
        Get the full configuration for this management zone short representation.
        """
        response = self._http_client.make_request(
            f"{ManagementZoneService.ENDPOINT}/{self.id}"
        ).json()
        return ManagementZone(http_client=self._http_client, raw_element=response)



from dynatrace.configuration_v1.geographic_regions import GeoRegionsIpAddressMappingsService
from dynatrace.http_client import HttpClient


class ConfigurationV1:
    def __init__(self, http_client: HttpClient):
        self._http_client = http_client

        self.geo_regions_ip_address_mappings = GeoRegionsIpAddressMappingsService(http_client)



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from enum import Enum
from datetime import datetime
from typing import List, Optional, Union, Dict, Any

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList
from dynatrace.utils import timestamp_to_string


class CustomTagService:
    ENDPOINT = "/api/v2/tags"

    def __init__(self, http_client: HttpClient) -> None:
        self.__http_client = http_client

    def list(
        self, entity_selector: str, time_from: Optional[Union[datetime, str]] = None, time_to: Optional[Union[datetime, str]] = None
    ) -> PaginatedList["METag"]:
        """
        Returns a list of custom tags
        :param entity_selector: specifies entities where you want to read tags
        :param time_from: The start of the requested timeframe.
        :param time_to: The end of the requested timeframe.

        :return: A list of METag objects
        """
        params = {"entitySelector": entity_selector, "from": timestamp_to_string(time_from), "to": timestamp_to_string(time_to)}

        return PaginatedList(METag, self.__http_client, target_url=self.ENDPOINT, target_params=params, list_item="tags")

    def post(
        self,
        entity_selector: str,
        tags: List["AddEntityTags"],
        time_from: Optional[Union[datetime, str]] = None,
        time_to: Optional[Union[datetime, str]] = None,
    ) -> "AddedEntityTags":
        """
        Adds custom tags to the specified entities
        :param entity_selector: specifies entities where you want to read tags
        :param time_from: The start of the requested timeframe.
        :param time_to: The end of the requested timeframe.
        :param tags: list of Tag objects Tag = { key, value(optional)}

        :return: AddedEntityTags
        """
        query_params = {
            "entitySelector": entity_selector,
            "from": timestamp_to_string(time_from),
            "to": timestamp_to_string(time_to),
        }
        body = {
            "tags": [t.to_json() for t in tags],
        }
        response = self.__http_client.make_request(self.ENDPOINT, params=body, method="POST", query_params=query_params).json()
        return AddedEntityTags(raw_element=response)

    def delete(
        self,
        key: str,
        entity_selector: str,
        value: Optional[str] = None,
        delete_all_with_key: Optional[bool] = None,
        time_from: Optional[Union[datetime, str]] = None,
        time_to: Optional[Union[datetime, str]] = None,
    ) -> "DeletedEntityTags":
        """
        Deletes the specified tag from the specified entities
        """
        params = {
            "key": key,
            "entitySelector": entity_selector,
            "deleteAllWithKey": delete_all_with_key,
            "value": value,
            "from": timestamp_to_string(time_from),
            "to": timestamp_to_string(time_to),
        }
        response = self.__http_client.make_request(self.ENDPOINT, params=params, method="DELETE")
        return DeletedEntityTags(raw_element=response.json())


class AddedEntityTags(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.matched_entities_count: int = raw_element.get("matchedEntitiesCount", 0)
        self.applied_tags: List[METag] = [METag(raw_element=tag) for tag in raw_element.get("appliedTags", [])]


class DeletedEntityTags(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.matched_entities_count: int = raw_element.get("matchedEntitiesCount", 0)


class AddEntityTags:
    def __init__(self, key: str, value: Optional[str] = None):
        self.key = key
        self.value = value

    def to_json(self) -> Dict[str, str]:
        return {"key": self.key, "value": self.value}


class METag(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.context: TagContext = TagContext(raw_element["context"])
        self.key: str = raw_element["key"]
        self.value: Optional[str] = raw_element.get("value")
        self.string_representation: Optional[str] = raw_element.get("stringRepresentation")

    def to_json(self) -> Dict[str, Any]:
        return {"context": str(self.context), "key": self.key, "value": self.value}


class TagContext(Enum):
    AWS = "AWS"
    AWS_GENERIC = "AWS_GENERIC"
    AZURE = "AZURE"
    CLOUD_FOUNDRY = "CLOUD_FOUNDRY"
    CONTEXTLESS = "CONTEXTLESS"
    ENVIRONMENT = "ENVIRONMENT"
    GOOGLE_CLOUD = "GOOGLE_CLOUD"
    KUBERNETES = "KUBERNETES"
    GOOGLE_COMPUTE_ENGINE = "GOOGLE_COMPUTE_ENGINE"

    def __str__(self) -> str:
        return self.value



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List
from enum import Enum
from requests import Response

from dynatrace.pagination import PaginatedList
from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient


class NotificationType(Enum):
    EMAIL = "EMAIL"
    PAGER_DUTY = "PAGER_DUTY"
    WEBHOOK = "WEBHOOK"
    SLACK = "SLACK"
    HIPCHAT = "HIPCHAT"
    VICTOROPS = "VICTOROPS"
    SERVICE_NOW = "SERVICE_NOW"
    XMATTERS = "XMATTERS"
    ANSIBLETOWER = "ANSIBLETOWER"
    OPS_GENIE = "OPS_GENIE"
    JIRA = "JIRA"
    TRELLO = "TRELLO"


class Notification(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))


class AnsibleTowerNotificationConfig(Notification):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))
        self.job_template_url: str = raw_element.get("jobTemplateURL")
        self.accept_any_certificate: bool = raw_element.get("acceptAnyCertificate")
        self.username: str = raw_element.get("username")
        self.password: str = raw_element.get("password")
        self.job_template_id: str = raw_element.get("jobTemplateID")
        self.custom_message: str = raw_element.get("customMessage")


class EmailNotificationConfig(Notification):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))
        self.subject: str = raw_element.get("subject")
        self.body: str = raw_element.get("body")
        self.receivers: List[str] = raw_element.get("receivers")
        self.cc_receivers: List[str] = raw_element.get("ccReceivers")
        self.bcc_receivers: List[str] = raw_element.get("bccReceivers")


class HipChatNotificationConfig(Notification):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))
        self.url: str = raw_element.get("url")
        self.message: str = raw_element.get("message")


class JiraNotificationConfig(Notification):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))
        self.username: str = raw_element.get("username")
        self.password: str = raw_element.get("password")
        self.url: str = raw_element.get("url")
        self.project_key: str = raw_element.get("projectKey")
        self.issue_type: str = raw_element.get("issueType")
        self.summary: str = raw_element.get("summary")
        self.description: str = raw_element.get("description")


class OpsGenieNotificationConfig(Notification):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))
        self.api_key: str = raw_element.get("apiKey")
        self.domain: str = raw_element.get("domain")
        self.message: str = raw_element.get("message")


class PagerDutyNotificationConfig(Notification):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))
        self.account: str = raw_element.get("account")
        self.service_api_key: str = raw_element.get("serviceApiKey")
        self.service_name: str = raw_element.get("serviceName")


class ServiceNowNotificationConfig(Notification):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))
        self.instance_name: str = raw_element.get("instanceName")
        self.url: str = raw_element.get("url")
        self.username: str = raw_element.get("username")
        self.password: str = raw_element.get("password")
        self.message: str = raw_element.get("message")
        self.send_incidents: bool = raw_element.get("sendIncidents")
        self.send_events: bool = raw_element.get("sendEvents")


class SlackNotificationConfig(Notification):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))
        self.url: str = raw_element.get("url")
        self.channel: str = raw_element.get("channel")
        self.title: str = raw_element.get("title")


class TrelloNotificationConfig(Notification):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))
        self.application_key: str = raw_element.get("applicationKey")
        self.authorization_token: str = raw_element.get("authorizationToken")
        self.board_id: str = raw_element.get("boardId")
        self.list_id: str = raw_element.get("listId")
        self.resolved_list_id: str = raw_element.get("resolvedListId")
        self.text: str = raw_element.get("text")
        self.description: str = raw_element.get("description")


class VictorOpsNotificationConfig(Notification):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))
        self.api_key: str = raw_element.get("apiKey")
        self.routing_key: str = raw_element.get("routingKey")
        self.message: str = raw_element.get("message")


class HttpHeader(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.name: str = raw_element.get("name")
        self.value: str = raw_element.get("value")


class WebHookNotificationConfig(Notification):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))
        self.url: str = raw_element.get("url")
        self.accept_any_certificate: bool = raw_element.get("acceptAnyCertificate")
        self.payload: str = raw_element.get("payload")
        self.headers: List[HttpHeader] = [HttpHeader(raw_element=header) for header in raw_element.get("headers", {})]
        self.notify_event_merges_enabled: bool = raw_element.get("notifyEventMergesEnabled")


class XMattersNotificationConfig(Notification):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.alerting_profile: str = raw_element.get("alertingProfile")
        self.active: bool = raw_element.get("active")
        self.type: NotificationType = NotificationType(raw_element.get("type"))
        self.url: str = raw_element.get("url")
        self.accept_any_certificate: bool = raw_element.get("acceptAnyCertificate")
        self.headers: List[HttpHeader] = [HttpHeader(raw_element=header) for header in raw_element.get("headers", {})]
        self.payload: str = raw_element.get("payload")


class NotificationConfigStub(DynatraceObject):
    def get_full_configuration(self) -> Notification:
        """
        Gets the full notification configuration for this stub.
        """
        response = self._http_client.make_request(f"/api/config/v1/notifications/{self.id}").json()
        if self.type == NotificationType.ANSIBLETOWER:
            notification = AnsibleTowerNotificationConfig(self._http_client, None, response)
        elif self.type == NotificationType.EMAIL:
            notification = EmailNotificationConfig(self._http_client, None, response)
        elif self.type == NotificationType.HIPCHAT:
            notification = HipChatNotificationConfig(self._http_client, None, response)
        elif self.type == NotificationType.JIRA:
            notification = JiraNotificationConfig(self._http_client, None, response)
        elif self.type == NotificationType.OPS_GENIE:
            notification = OpsGenieNotificationConfig(self._http_client, None, response)
        elif self.type == NotificationType.PAGER_DUTY:
            notification = PagerDutyNotificationConfig(self._http_client, None, response)
        elif self.type == NotificationType.SERVICE_NOW:
            notification = ServiceNowNotificationConfig(self._http_client, None, response)
        elif self.type == NotificationType.SLACK:
            notification = SlackNotificationConfig(self._http_client, None, response)
        elif self.type == NotificationType.TRELLO:
            notification = TrelloNotificationConfig(self._http_client, None, response)
        elif self.type == NotificationType.VICTOROPS:
            notification = VictorOpsNotificationConfig(self._http_client, None, response)
        elif self.type == NotificationType.WEBHOOK:
            notification = WebHookNotificationConfig(self._http_client, None, response)
        elif self.type == NotificationType.XMATTERS:
            notification = XMattersNotificationConfig(self._http_client, None, response)
        else:
            notification = Notification(self._http_client, None, response)
        return notification

    def delete(self) -> Response:
        """
        Delete the notification for this stub.
        """
        return self._http_client.make_request(f"/api/config/v1/notifications/{self.id}", method="DELETE")

    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.description: str = raw_element.get("description")
        self.type: NotificationType = NotificationType(raw_element.get("type"))


class NotificationService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(self) -> PaginatedList[NotificationConfigStub]:
        """
        Lists all alerting profiles in the environmemt. No configurable parameters.
        """
        return PaginatedList(NotificationConfigStub, self.__http_client, f"/api/config/v1/notifications", list_item="values")

    def delete(self, notification_id: str) -> Response:
        """
        Delete the notification with the specified id.
        """
        return self.__http_client.make_request(f"/api/config/v1/notifications/{notification_id}", method="DELETE")



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Generic, TypeVar, Iterator, TYPE_CHECKING

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient

T = TypeVar("T", bound=DynatraceObject)


class PaginatedList(Generic[T]):
    def __init__(self, target_class, http_client, target_url, target_params=None, headers=None, list_item="result"):
        self.__target_class = target_class
        self.__http_client: HttpClient = http_client
        self.__target_url = target_url
        self.__target_params = target_params
        self.__headers = headers
        self.__list_item = list_item
        self._has_next_page = True
        self.__total_count = None
        self.__page_size = None

        self.__elements = self._get_next_page()

    def __getitem__(self, index):
        pass

    def __iter__(self) -> Iterator[T]:
        for element in self.__elements:
            yield element

        while self._has_next_page:
            new_elements = self._get_next_page()
            for element in new_elements:
                yield element

    def __len__(self):
        return self.__total_count or len(self.__elements)

    def _get_next_page(self):
        response = self.__http_client.make_request(self.__target_url, params=self.__target_params, headers=self.__headers)
        json_response = response.json()
        data = []
        if json_response.get("nextPageKey", None):
            self._has_next_page = True
            self.__target_params = {"nextPageKey": json_response["nextPageKey"]}
        else:
            self._has_next_page = False

        if self.__list_item in json_response:
            elements = json_response[self.__list_item]
            self.__total_count = json_response.get("totalCount") or len(elements)

            data = [self.__target_class(self.__http_client, response.headers, element) for element in elements]
        return data


class HeaderPaginatedList(Generic[T]):
    def __init__(self, target_class, http_client, target_url, target_params=None, headers=None):
        self.__elements = list()
        self.__target_class = target_class
        self.__http_client: HttpClient = http_client
        self.__target_url = target_url
        self.__target_params = target_params
        self.__headers = headers
        self._has_next_page = True
        self.__total_count = None
        self.__page_size = None

    def __getitem__(self, index):
        pass

    def __iter__(self) -> Iterator[T]:
        for element in self.__elements:
            yield element

        while self._has_next_page:
            new_elements = self._get_next_page()
            for element in new_elements:
                yield element

    def __len__(self):
        return self.__total_count or len(self.__elements)

    def _get_next_page(self):
        response = self.__http_client.make_request(self.__target_url, params=self.__target_params, headers=self.__headers)
        json_response = response.json()
        headers = response.headers
        if "next-page-key" in headers:
            self._has_next_page = True
            self.__target_params = {"nextPageKey": headers["next-page-key"]}
        else:
            self._has_next_page = False

        elements = json_response
        self.__total_count = headers.get("total-count") or len(elements)
        data = [self.__target_class(self.__http_client, response.headers, element) for element in elements]
        return data



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from enum import Enum
from typing import Dict, Any, Union, List

from requests import Response
from datetime import datetime
from typing import Optional, Union, Dict, Any, List

from dynatrace.http_client import HttpClient
from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.pagination import PaginatedList
from dynatrace.utils import timestamp_to_string


class LogService:
    # TODO - Add search and aggregate
    ENDPOINT = "/api/v2/logs"

    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def export(
        self,
        query: Optional[str] = None,
        time_from: Optional[Union[datetime, str]] = None,
        time_to: Optional[Union[datetime, str]] = None,
        sort: Optional[str] = None,
        page_size: Optional[int] = None,
    ) -> PaginatedList["LogRecord"]:
        """
        Gets the log records matching the provided criteria. Retrieves all records using pagination.
        :param query: The log search query
        :param page_size: Number of results per page
        :param time_from: Start of the requested timeframe
        :param time_to: End of the requested timefram
        :param sort: Defines the ordering of log records
        :return A list of log records
        """
        params = {
            "query": query,
            "pageSize": page_size,
            "from": timestamp_to_string(time_from),
            "to": timestamp_to_string(time_to),
            "sort": sort,
        }
        return PaginatedList(LogRecord, self.__http_client, "/api/v2/logs/export", params, list_item="results")

    def ingest(self, payload: Union[Dict[str, Any], List[Dict[str, Any]]]) -> Response:
        """
        Ingests logs into the Dynatrace log store.
        :param payload: A list of log entries or a single log entry, which are JSON objects (dictionaries)
        :return: The HTTP Response
        """
        headers = {"Content-Type": "application/json; charset=utf-8"}
        return self.__http_client.make_request(f"{self.ENDPOINT}/ingest", params=payload, method="POST", headers=headers)
    
class LogRecord(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.additional_columns: dict = raw_element.get("additionalColumns")
        self.event_type: EventType = EventType(raw_element.get("eventType"))
        self.timestamp: datetime = datetime.utcfromtimestamp(raw_element.get("timestamp") / 1000)
        self.content: str = raw_element.get("content")
        self.status: LogRecordStatus = LogRecordStatus(raw_element.get("status"))

class EventType(Enum):
    K8S = "K8S"
    LOG = "LOG"
    SFM = "SFM"

class LogRecordStatus(Enum):
    ERROR = "ERROR"
    INFO = "INFO"
    NONE = "NONE"
    NOT_APPLICABLE = "NOT_APPLICABLE"
    WARN = "WARN"


"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import defaultdict
from datetime import datetime
from typing import List, Optional, Dict

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient

SYNTHETIC_EVENT_TYPE_OUTAGE = "testOutage"
SYNTHETIC_EVENT_TYPE_SLOWDOWN = "testSlowdown"


class ThirdPartySyntheticTestsService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def report_simple_thirdparty_synthetic_test(
        self,
        engine_name: str,
        timestamp: datetime,
        location_id: str,
        location_name: str,
        test_id: str,
        test_title: str,
        schedule_interval: int,
        success: bool,
        response_time: int,
        icon_url: str = None,
        edit_link: str = None,
        step_title: Optional[str] = None,
        detailed_steps: Optional[List["SyntheticTestStep"]] = None,
        detailed_step_results: Optional[List["SyntheticMonitorStepResult"]] = None,
    ):

        location = ThirdPartySyntheticLocation(self.__http_client, location_id, location_name)
        synthetic_location = SyntheticTestLocation(self.__http_client, location_id)
        if detailed_steps is None:
            detailed_steps = [SyntheticTestStep(self.__http_client, 1, step_title)]

        monitor = ThirdPartySyntheticMonitor(
            self.__http_client,
            test_id,
            test_title,
            [synthetic_location],
            schedule_interval,
            steps=detailed_steps,
            edit_link=edit_link,
        )
        if detailed_step_results is None:
            detailed_step_results = [SyntheticMonitorStepResult(self.__http_client, 1, timestamp, response_time_millis=response_time)]
        location_result = ThirdPartySyntheticLocationTestResult(self.__http_client, location_id, timestamp, success, step_results=detailed_step_results)
        test_result = ThirdPartySyntheticResult(self.__http_client, test_id, len(detailed_steps), [location_result])
        tests = ThirdPartySyntheticTests(self.__http_client, engine_name, timestamp, [location], [monitor], [test_result], synthetic_engine_icon_url=icon_url)
        return tests.post()

    def create_synthetic_test_step_result(self, step_id: int, timestamp: datetime, response_time: int) -> "SyntheticMonitorStepResult":
        return SyntheticMonitorStepResult(self.__http_client, step_id, timestamp, response_time_millis=response_time)

    def create_synthetic_test_step(self, step_id: int, step_title: str) -> "SyntheticTestStep":
        return SyntheticTestStep(self.__http_client, step_id, step_title)

    def report_simple_thirdparty_synthetic_test_event(
        self,
        test_id: str,
        name: str,
        location_id: str,
        timestamp: datetime,
        state: str,
        event_type: str,
        reason: str,
        engine_name: str,
    ):
        opened_events: List[ThirdPartyEventOpenNotification] = []
        resolved_events = []
        event_id = f"{test_id}_event"
        if state == "open":
            opened_events.append(ThirdPartyEventOpenNotification(self.__http_client, test_id, event_id, name, event_type, reason, timestamp, [location_id]))
        else:
            resolved_events.append(ThirdPartyEventResolvedNotification(self.__http_client, test_id, event_id, timestamp))

        if opened_events or resolved_events:
            events = ThirdPartySyntheticEvents(self.__http_client, engine_name, opened_events, resolved_events)
            return events.post()


class ThirdPartySyntheticTests(DynatraceObject):
    def __init__(
        self,
        http_client,
        synthetic_engine_name: str,
        message_timestamp: datetime,
        locations: List["ThirdPartySyntheticLocation"],
        tests: List["ThirdPartySyntheticMonitor"],
        test_results: Optional[List["ThirdPartySyntheticResult"]] = None,
        synthetic_engine_icon_url: Optional[str] = None,
    ):

        raw_element = {
            "syntheticEngineName": synthetic_engine_name,
            "syntheticEngineIconUrl": synthetic_engine_icon_url,
            "messageTimestamp": int(message_timestamp.timestamp() * 1000),
            "locations": [location._raw_element for location in locations],
            "tests": [test._raw_element for test in tests],
            "testResults": [test_result._raw_element for test_result in test_results] if test_results else None,
        }
        super().__init__(http_client, None, raw_element)

    def post(self):
        return self._http_client.make_request(f"/api/v1/synthetic/ext/tests", params=self._raw_element, method="POST")


class ThirdPartySyntheticLocation(DynatraceObject):
    def __init__(self, http_client, location_id: str, name: str, ip: Optional[str] = None):

        raw_element = {"id": location_id, "name": name, "ip": ip}
        super().__init__(http_client, None, raw_element)


class ThirdPartySyntheticMonitor(DynatraceObject):
    def __init__(
        self,
        http_client,
        test_id: str,
        title: str,
        locations: List["SyntheticTestLocation"],
        schedule_interval_in_seconds: int,
        description: Optional[str] = None,
        test_setup: Optional[str] = None,
        expiration_timestamp: Optional[int] = None,
        drilldown_link: Optional[str] = None,
        edit_link: Optional[str] = None,
        deleted: Optional[bool] = None,
        steps: Optional[List["SyntheticTestStep"]] = None,
        no_data_timeout: Optional[int] = None,
    ):

        raw_element = {
            "id": test_id,
            "title": title,
            "description": description,
            "testSetup": test_setup,
            "expirationTimestamp": expiration_timestamp,
            "drilldownLink": drilldown_link,
            "editLink": edit_link,
            "deleted": deleted,
            "locations": [location._raw_element for location in locations],
            "steps": [step._raw_element for step in steps] if steps else None,
            "scheduleIntervalInSeconds": schedule_interval_in_seconds,
            "noDataTimeout": no_data_timeout,
        }
        super().__init__(http_client, None, raw_element)


class SyntheticTestLocation(DynatraceObject):
    def __init__(self, http_client, location_id: str, enabled: Optional[bool] = None):

        raw_element = {"id": location_id, "enabled": enabled}
        super().__init__(http_client, None, raw_element)


class SyntheticTestStep(DynatraceObject):
    def __init__(self, http_client, step_id: int, title: str):
        self.step_id = step_id
        self.title = title

        raw_element = {"id": step_id, "title": title}
        super().__init__(http_client, None, raw_element)


class ThirdPartySyntheticResult(DynatraceObject):
    def __init__(
        self,
        http_client,
        test_id: str,
        total_step_count: int,
        location_results: List["ThirdPartySyntheticLocationTestResult"],
        schedule_interval_in_seconds: Optional[int] = None,
    ):

        raw_element = {
            "id": test_id,
            "scheduleIntervalInSeconds": schedule_interval_in_seconds,
            "totalStepCount": total_step_count,
            "locationResults": [location_result._raw_element for location_result in location_results],
        }
        super().__init__(http_client, None, raw_element)


class ThirdPartySyntheticLocationTestResult(DynatraceObject):
    def __init__(
        self,
        http_client,
        location_id: str,
        start_timestamp: datetime,
        success: bool,
        success_rate: Optional[float] = None,
        response_time_millis: Optional[int] = None,
        step_results: Optional[List["SyntheticMonitorStepResult"]] = None,
    ):

        raw_element = {
            "id": location_id,
            "startTimestamp": int(start_timestamp.timestamp() * 1000),
            "successRate": success_rate,
            "success": success,
            "responseTimeMillis": response_time_millis,
            "stepResults": [step_result._raw_element for step_result in step_results] if step_results else None,
        }
        super().__init__(http_client, None, raw_element)


class SyntheticMonitorStepResult(DynatraceObject):
    def __init__(
        self,
        http_client,
        step_id: int,
        start_timestamp: datetime,
        response_time_millis: Optional[int] = None,
        error: Optional["SyntheticMonitorError"] = None,
    ):

        raw_element = {
            "id": step_id,
            "startTimestamp": int(start_timestamp.timestamp() * 1000),
            "responseTimeMillis": response_time_millis,
            "error": error._raw_element if error else None,
        }
        super().__init__(http_client, None, raw_element)


class SyntheticMonitorError(DynatraceObject):
    def __init__(self, http_client, code: int, message: str):

        raw_element = {"code": code, "message": message}
        super().__init__(http_client, None, raw_element)


class ThirdPartySyntheticEvents(DynatraceObject):
    def __init__(
        self,
        http_client,
        synthetic_engine_name: str,
        open_events: Optional[List["ThirdPartyEventOpenNotification"]],
        resolved_events: Optional[List["ThirdPartyEventResolvedNotification"]],
    ):

        raw_element = {
            "syntheticEngineName": synthetic_engine_name,
            "open": [open_event._raw_element for open_event in open_events] if open_events else None,
            "resolved": [resolved_event._raw_element for resolved_event in resolved_events] if resolved_events else None,
        }
        super().__init__(http_client, None, raw_element)

    def post(self):
        return self._http_client.make_request(f"/api/v1/synthetic/ext/events", params=self._raw_element, method="POST")


class ThirdPartyEventOpenNotification(DynatraceObject):
    def __init__(
        self,
        http_client,
        test_id: str,
        event_id: str,
        name: str,
        event_type: str,
        reason: str,
        start_timestamp: datetime,
        location_ids: List[str],
    ):

        raw_element = {
            "testId": test_id,
            "eventId": event_id,
            "name": name,
            "eventType": event_type,
            "reason": reason,
            "startTimestamp": int(start_timestamp.timestamp() * 1000),
            "locationIds": location_ids,
        }
        super().__init__(http_client, None, raw_element)


class ThirdPartyEventResolvedNotification(DynatraceObject):
    def __init__(self, http_client, test_id: str, event_id: str, end_timestamp: datetime):

        raw_element = {"testId": test_id, "eventId": event_id, "endTimestamp": int(end_timestamp.timestamp() * 1000)}
        super().__init__(http_client, None, raw_element)



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from enum import Enum
from datetime import datetime
from requests import Response
from typing import Dict, Any, Optional, Union

from dynatrace.utils import timestamp_to_string
from dynatrace.http_client import HttpClient
from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.pagination import PaginatedList


class SloService:
    ENDPOINT = "/api/v2/slo"

    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(
        self,
        page_size: Optional[int] = 10,
        time_from: Optional[Union[datetime, str]] = "now-2w",
        time_to: Optional[Union[datetime, str]] = None,
        slo_selector: Optional[str] = None,
        sort: Optional[str] = "name",
        time_frame: Optional[str] = "CURRENT",
        page_idx: Optional[int] = 1,
        demo: Optional[bool] = False,
        evaluate: Optional[str] = "false",
        enabled_slos: Optional[str] = "all"
    ) -> PaginatedList["Slo"]:
        """Lists all available SLOs along with calculated values

        :param page_size: The amount of SLOs in a single response payload. Max 10,000.
        :param time_from: The start of the requested timeframe.
        :param time_to: The end of the requested timeframe.
        :param slo_selector: The scope of the query. Only SLOs matching the provided criteria are included in the response.
        :param sort: The sorting of SLO entries. Default is by name in ascending order.
        :param time_frame: The timeframe to calculate the SLO values. CURRENT: SLO's own timeframe. GTF: timeframe specified by from and to parameters.
        :param page_idx: Only SLOs on the given page are included in the response. The first page has the index '1'.
        :param demo: Get your SLOs (false) or a set of demo SLOs (true)
        :param evaluate: Get your SLOs without them being evaluated ("false") or with evaluations ("true"). This value must be a lowercase string.
        :param enabled_slos: Get your enabled SLOs ("true"), disabled ones ("false") or both enabled and disabled ones ("all"). This value must be a lowercase string.

        :returns PaginatedList[Slo]: the list of SLOs matching criteria
        """
        params = {
            "pageSize": page_size,
            "from": timestamp_to_string(time_from),
            "to": timestamp_to_string(time_to),
            "sloSelector": slo_selector,
            "sort": sort,
            "timeFrame": time_frame,
            "pageIdx": page_idx,
            "demo": demo,
            "evaluate": evaluate,
            "enabledSlos": enabled_slos
        }
        return PaginatedList(target_class=Slo, http_client=self.__http_client, target_params=params, target_url=f"{self.ENDPOINT}", list_item="slo")

    def get(
        self, 
        slo_id: str, 
        time_from: Optional[Union[datetime, str]] = "now-2w", 
        time_to: Optional[Union[datetime, str]] = None, 
        time_frame: Optional[str] = "CURRENT"
    ) -> "Slo":
        """Gets parameters and the calculated value of an SLO

        :param slo_id: The ID of the required SLO.
        :param time_from: The start of the requested timeframe.
        :param time_to: The end of the requested timeframe.
        :param time_frame: The timeframe to calculate the SLO values. CURRENT: SLO's own timeframe. GTF: timeframe specified by from and to parameters.

        :returns Slo: the requested SLO
        """
        params = {
            "from": timestamp_to_string(time_from),
            "to": timestamp_to_string(time_to),
            "timeFrame": time_frame,
        }
        response = self.__http_client.make_request(f"{self.ENDPOINT}/{slo_id}", params=params).json()
        return Slo(raw_element=response)

    def post(self, slo: "Slo") -> "Response":
        """Creates a new SLO in Dynatrace.

        :param slo: the Slo that should be created.

        :returns Response: HTTP response for the request
        """
        return slo.post()

    def put(self, slo: "Slo") -> "Response":
        """Updates an existing SLO in Dynatrace.

        :param slo: the Slo with udpated details

        :returns Response: HTTP response for the request
        """
        return slo.put()

    def delete(self, slo_id: str) -> "Response":
        """Deletes an SLO

        :param slo_id: The ID of the existing SLO.

        :returns Response: HTTP response for the request
        """
        return self.__http_client.make_request(path=f"{self.ENDPOINT}/{slo_id}", method="DELETE")

    def create(
        self,
        name: str,
        target: float,
        warning: float,
        timeframe: str,
        use_rate_metric: Optional[bool] = None,
        metric_rate: Optional[str] = None,
        metric_numerator: Optional[str] = None,
        metric_denominator: Optional[str] = None,
        metric_expression: Optional[str] = None,
        metric_name: Optional[str] = None,
        filter: Optional[str] = None,
        evaluation_type: Optional[str] = "AGGREGATE",
        custom_description: Optional[str] = None,
        enabled: Optional[bool] = False,
    ) -> "Slo":
        """Creates an Slo object from scratch.

        :param name: The name of the SLO.
        :param target: The target value of the SLO.
        :param warning: The warning value of the SLO. At warning state the SLO is still fulfilled but is getting close to failure.
        :param timeframe: The timeframe for the SLO evaluation. Use the syntax of the global timeframe selector.
        :param use_rate_metric: [DEPRECATED] The type of the metric to use for SLO calculation - an existing percentage-based metric (true) or a ratio of two metrics (false)
        :param metric_rate: [DEPRECATED] The percentage-based metric for the calculation of the SLO. Required when the useRateMetric is set to true.
        :param metric_numerator: [DEPRECATED] The metric for the count of successes (the numerator in rate calculation).Required when the useRateMetric is set to false.
        :param metric_denominator: [DEPRECATED] The total count metric (the denominator in rate calculation). Required when the useRateMetric is set to false.
        :param metric_expression: The percentage-based metric expression for the calculation of the SLO.
        :param evaluation_type: The evaluation type of the SLO.
        :param filter_: The entity filter for the SLO evaluation. Use the syntax of entity selector.
        :param custom_description: The custom description of the SLO.
        :param enabled: The SLO is enabled (true) or disabled (false).

        :returns Slo: the resulting Slo. This can now be used in POST and PUT calls
        """
        raw_slo = {
            "name": name,
            "target": target,
            "warning": warning,
            "timeframe": timeframe,
            "useRateMetric": use_rate_metric,
            "metricRate": metric_rate if use_rate_metric else "",
            "metricNumerator": metric_numerator if not use_rate_metric else "",
            "metricDenominator": metric_denominator if not use_rate_metric else "",
            "metricExpression": metric_expression,
            "metricName": metric_name,
            "filter": filter,
            "evaluationType": evaluation_type,
            "description": custom_description,
            "enabled": enabled,
        }
        return Slo(raw_element=raw_slo, http_client=self.__http_client)


class Slo(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        # required
        self.name: str = raw_element.get("name")
        self.id: str = raw_element.get("id", "")
        self.target: float = raw_element.get("target")
        self.warning: float = raw_element.get("warning")
        self.timeframe: str = raw_element.get("timeframe")
        self.evaluation_type: SloEvaluationType = SloEvaluationType(raw_element.get("evaluationType"))

        # optional
        self.status: Optional[SloStatus] = SloStatus(raw_element.get("status")) if raw_element.get("status") else None
        self.metric_rate: Optional[str] = raw_element.get("metricRate")
        self.metric_numerator: Optional[str] = raw_element.get("metricNumerator")
        self.metric_denominator: Optional[str] = raw_element.get("metricDenominator")
        self.metric_expression: Optional[str] = raw_element.get("metricExpression")
        self.metric_name: Optional[str] = raw_element.get("metricName")
        self.error_budget: Optional[float] = raw_element.get("errorBudget", 0)
        self.numerator_value: Optional[float] = raw_element.get("numeratorValue", 0)
        self.denominator_value: Optional[float] = raw_element.get("denominatorValue", 0)
        self.related_open_problems: Optional[int] = raw_element.get("relatedOpenProblems", 0)
        self.evaluated_percentage: Optional[float] = raw_element.get("evaluatedPercentage", 0)
        self.filter: Optional[str] = raw_element.get("filter")
        self.enabled: Optional[bool] = raw_element.get("enabled", False)
        self.custom_description: Optional[str] = raw_element.get("description")
        self.error: Optional[SloError] = SloError(raw_element.get("error", SloError.NONE))
        self.use_rate_metric: Optional[bool] = raw_element.get("useRateMetric")

    def to_json(self) -> Dict[str, Any]:
        """Translates an Slo to a JSON dict."""
        return {
            "name": self.name,
            "target": self.target,
            "warning": self.warning,
            "timeframe": self.timeframe,
            "evaluationType": str(self.evaluation_type),
            "enabled": self.enabled,
            "useRateMetric": self.use_rate_metric,
            "metricRate": self.metric_rate,
            "metricNumerator": self.metric_numerator,
            "metricDenominator": self.metric_denominator,
            "metricExpression": self.metric_expression,
            "metricName": self.metric_name,
            "filter": self.filter,
            "customDescription": self.custom_description,
        }

    def post(self) -> "Response":
        """Creates this object as a new SLO in Dynatrace"""
        response = self._http_client.make_request(path=SloService.ENDPOINT, method="POST", params=self.to_json())
        if response.status_code == 201:
            location = response.headers.get("location")
            self.id = location[location.rfind("/") + 1 :].strip()

        return response

    def put(self) -> "Response":
        """Updates an existing SLO in Dynatrace based on this object's details"""
        return self._http_client.make_request(path=f"{SloService.ENDPOINT}/{self.id}", method="PUT", params=self.to_json())


class SloEvaluationType(Enum):
    AGGREGATE = "AGGREGATE"

    def __str__(self):
        return self.value


class SloStatus(Enum):
    FAILURE = "FAILURE"
    SUCCESS = "SUCCESS"
    WARNING = "WARNING"

    def __str__(self):
        return self.value


class SloError(Enum):
    DIFFERENT_ENTITY_TYPE_IN_METRICS = "DIFFERENT_ENTITY_TYPE_IN_METRICS"
    EVALUATION_TIMEFRAME_OUT_OF_BOUNDS = "EVALUATION_TIMEFRAME_OUT_OF_BOUNDS"
    FILTER_MATCHES_IN_CONDITION_LIMIT_EXCEEDED = "FILTER_MATCHES_IN_CONDITION_LIMIT_EXCEEDED"
    INTERNAL_ERROR = "INTERNAL_ERROR"
    INVALID_ENTITY_SELECTOR = "INVALID_ENTITY_SELECTOR"
    INVALID_METRIC_DENOMINATOR = "INVALID_METRIC_DENOMINATOR"
    INVALID_METRIC_NUMERATOR = "INVALID_METRIC_NUMERATOR"
    INVALID_METRIC_SELECTOR = "INVALID_METRIC_SELECTOR"
    INVALID_TIMEFRAME = "INVALID_TIMEFRAME"
    METRICS_NOT_RESOLVED = "METRICS_NOT_RESOLVED"
    METRICS_NO_DATA = "METRICS_NO_DATA"
    METRIC_DENOMINATOR_NOT_RESOLVED = "METRIC_DENOMINATOR_NOT_RESOLVED"
    METRIC_DENOMINATOR_NO_DATA = "METRIC_DENOMINATOR_NO_DATA"
    METRIC_DENOMINATOR_NO_DATA_POINTS = "METRIC_DENOMINATOR_NO_DATA_POINTS"
    METRIC_DENOMINATOR_ZERO = "METRIC_DENOMINATOR_ZERO"
    METRIC_EXPRESSION_NOT_RESOLVED = "METRIC_EXPRESSION_NOT_RESOLVED"
    METRIC_EXPRESSION_NO_DATA = "METRIC_EXPRESSION_NO_DATA"
    METRIC_EXPRESSION_NO_DATA_POINTS = "METRIC_EXPRESSION_NO_DATA_POINTS"
    METRIC_GENERIC_TSM_FAULT = "METRIC_GENERIC_TSM_FAULT"
    METRIC_NUMERATOR_NOT_RESOLVED = "METRIC_NUMERATOR_NOT_RESOLVED"
    METRIC_NUMERATOR_NO_DATA = "METRIC_NUMERATOR_NO_DATA"
    METRIC_NUMERATOR_NO_DATA_POINTS = "METRIC_NUMERATOR_NO_DATA_POINTS"
    METRIC_RATE_NOT_RESOLVED = "METRIC_RATE_NOT_RESOLVED"
    METRIC_RATE_NO_DATA = "METRIC_RATE_NO_DATA"
    METRIC_RATE_NO_DATA_POINTS = "METRIC_RATE_NO_DATA_POINTS"
    METRIC_TOO_MANY_RESULTS = "METRIC_TOO_MANY_RESULTS"
    NONE = "NONE"

    def __str__(self):
        return self.value



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Optional, Dict, Any, Union
from enum import Enum

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient


class Method(Enum):
    OFF = "OFF"
    MINIMUM_THRESHOLD = "MINIMUM_THRESHOLD"
    PROCESS_IMPACT = "PROCESS_IMPACT"


class AvailabilityMonitoringPG(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.method: Method = Method(raw_element.get("method"))
        self.minimum_threshold: int = raw_element.get("minimumThreshold")


class AnomalyDetectionPG(DynatraceObject):
    @staticmethod
    def create(method: Union[str, Method], minimum_threshold: Optional[int] = None):
        method = Method(method)
        if method == Method.MINIMUM_THRESHOLD:
            raw_element = {"availabilityMonitoring": {"method": method.value, "minimumThreshold": minimum_threshold}}
        else:
            raw_element = {"availabilityMonitoring": {"method": method.value}}
        return AnomalyDetectionPG(raw_element=raw_element)

    def json(self):
        """
        Get the json representation of this process group anomaly detection config.
        """
        if self.availability_monitoring.method == Method.MINIMUM_THRESHOLD:
            json = {
                "availabilityMonitoring": {
                    "method": self.availability_monitoring.method.value,
                    "minimumThreshold": self.availability_monitoring.minimum_threshold,
                }
            }
        else:
            json = {"availabilityMonitoring": {"method": self.availability_monitoring.method.value}}
        return json

    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.availability_monitoring: AvailabilityMonitoringPG = AvailabilityMonitoringPG(
            self._http_client, raw_element=raw_element.get("availabilityMonitoring")
        )


class AnomalyDetectionPGService:
    # TODO - Early adopter endpoint, be sure to check back in for updates
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def get_configuration(self, id: str):
        """
        Get the anomaly detection configuration for the specified process group.
        """
        return AnomalyDetectionPG(self.__http_client, raw_element=self.__http_client.make_request(f"/api/config/v1/anomalyDetection/processGroups/{id}").json())

    def put_configuration(self, id: str, anomaly_detection_config: AnomalyDetectionPG):
        """
        Update the anomaly detection configuration for the specified process group.
        """
        return self.__http_client.make_request(f"/api/config/v1/anomalyDetection/processGroups/{id}", method="PUT", params=anomaly_detection_config.json())

    def delete_configuration(self, id: str):
        """
        Deletes the anomaly detection configuration for the process group. This is the same as turning availaility monitoring off.
        """
        return self.__http_client.make_request(f"/api/config/v1/anomalyDetection/processGroups/{id}", method="DELETE")



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from requests import Response
from dynatrace.environment_v2.monitored_entities import EntityShortRepresentation
from dynatrace.configuration_v1.endpoint import EndpointShortRepresentation
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList
from dynatrace.dynatrace_object import DynatraceObject


class PluginService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(self) -> PaginatedList["PluginShortRepresentation"]:
        """
        List all uploaded plugins
        """
        return PaginatedList(PluginShortRepresentation, self.__http_client, "/api/config/v1/plugins", list_item="values")

    def list_states(self, plugin_id) -> PaginatedList["PluginState"]:
        """
        List the states of the specified plugin
        """
        return PaginatedList(PluginState, self.__http_client, f"/api/config/v1/plugins/{plugin_id}/states", list_item="states")

    def delete(self, plugin_id) -> Response:
        """
        Deletes the ZIP file of the specified plugin
        :param plugin_id: The ID of the plugin to be deleted
        """
        return self.__http_client.make_request(f"/api/config/v1/plugins/{plugin_id}/binary", method="DELETE")


class PluginState(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.plugin_id: str = raw_element.get("pluginId")
        self.version: str = raw_element.get("version")
        self.endpoint_id: str = raw_element.get("endpointId")
        self.state: str = raw_element.get("state")
        self.state_description: str = raw_element.get("stateDescription")
        self.timestamp: str = raw_element.get("timestamp")
        self.host_id: str = raw_element.get("hostId")
        self.process_id: str = raw_element.get("processId")


class PluginShortRepresentation(EntityShortRepresentation):
    def delete(self) -> Response:
        """
        Deletes the ZIP file of this plugin
        """
        return self._http_client.make_request(f"/api/config/v1/plugins/{self.id}/binary", method="DELETE")

    @property
    def endpoints(self) -> PaginatedList[EndpointShortRepresentation]:
        return PaginatedList(
            EndpointShortRepresentation,
            self._http_client,
            f"/api/config/v1/plugins/{self.id}/endpoints",
            None,
            list_item="values",
        )

    @property
    def states(self) -> PaginatedList[PluginState]:
        return PaginatedList(PluginState, self._http_client, f"/api/config/v1/plugins/{self.id}/states", None, list_item="states")



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Optional, Dict, Any, Union
from enum import Enum

from dynatrace.pagination import PaginatedList
from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient


class MonitorType(Enum):
    BROWSER = "BROWSER"
    HTTP = "HTTP"


class LoadingTimeThresholdType(Enum):
    ACTION = "ACTION"
    TOTAL = "TOTAL"


class TagSource(Enum):
    AUTO = "AUTO"
    RULE_BASED = "RULE_BASED"
    USER = "USER"


class CreatedFrom(Enum):
    API = "API"
    GUI = "GUI"


class TagContext(Enum):
    AWS = "AWS"
    AWS_GENERIC = "AWS_GENERIC"
    AZURE = "AZURE"
    CLOUD_FOUNDRY = "CLOUD_FOUNDRY"
    CONTEXTLESS = "CONTEXTLESS"
    ENVIRONMENT = "ENVIRONMENT"
    GOOGLE_CLOUD = "GOOGLE_CLOUD"
    KUBERNETES = "KUBERNETES"


class MonitorCollectionElement(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.name: str = raw_element.get("name")
        self.entity_id: str = raw_element.get("entityId")
        self.monitor_type: str = raw_element.get("type")
        self.enabled: bool = raw_element.get("enabled")


class LocalOutagePolicy(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.affected_locations: int = raw_element.get("affectedLocations")
        self.consecutive_runs: int = raw_element.get("consecutiveRuns")


class OutageHandlingPolicy(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.global_outage: bool = raw_element.get("globalOutage")
        self.local_outage: bool = raw_element.get("localOutage")
        self.local_outage_policy: LocalOutagePolicy = LocalOutagePolicy(raw_element=raw_element.get("localOutagePolicy"))
        self.retry_on_error: bool = raw_element.get("retryOnError")


class LoadingTimeThreshold(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.type: LoadingTimeThresholdType = LoadingTimeThresholdType(raw_element.get("type"))
        self.value_ms: int = raw_element.get("valueMs")
        self.request_index: int = raw_element.get("requestIndex")
        self.event_index: int = raw_element.get("eventIndex")


class LoadingTimeThresholdsPolicyDto(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.enabled: bool = raw_element.get("enabled")
        self.thresholds: List[LoadingTimeThreshold] = [LoadingTimeThreshold(raw_element=threshold) for threshold in raw_element.get("thresholds")]


class AnomalyDetection(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.outage_handling: OutageHandlingPolicy = OutageHandlingPolicy(raw_element=raw_element.get("outageHandling"))
        self.loading_time_thresholds: LoadingTimeThresholdsPolicyDto = LoadingTimeThresholdsPolicyDto(raw_element=raw_element.get("loadingTimeThresholds"))


class TagWithSourceInfo(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.source: TagSource = TagSource(raw_element.get("source"))
        self.context: TagContext = TagContext(raw_element.get("context"))
        self.key: str = raw_element.get("key")
        self.value: str = raw_element.get("value")


class ManagementZone(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")


class SyntheticMonitor(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.entity_id: str = raw_element.get("entityId")
        self.name: str = raw_element.get("name")
        self.frequency_min: int = raw_element.get("frequencyMin")
        self.enabled: bool = raw_element.get("enabled")
        self.type: MonitorType = MonitorType(raw_element.get("type"))
        self.created_from: CreatedFrom = CreatedFrom(raw_element.get("createdFrom"))
        self.script: dict = raw_element.get("script")
        self.locations: List[str] = raw_element.get("locations")
        self.anomaly_detection: AnomalyDetection = AnomalyDetection(raw_element=raw_element.get("anomalyDetection"))
        self.tags: List[TagWithSourceInfo] = [TagWithSourceInfo(raw_element=tag) for tag in raw_element.get("tags")]
        self.management_zones: List[ManagementZone] = [ManagementZone(raw_element=zone) for zone in raw_element.get("managementZones")]
        self.automatically_assigned_apps: List[str] = raw_element.get("automaticallyAssignedApps")
        self.manually_assigned_apps: List[str] = raw_element.get("manuallyAssignedApps")


class SyntheticMonitorsService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(self, monitor_type: Optional[Union[MonitorType, str]] = None) -> PaginatedList[MonitorCollectionElement]:
        """
        Lists all synthetic monitors in the environment.
        """
        params = {"type": MonitorType(monitor_type).value if monitor_type else None}
        return PaginatedList(MonitorCollectionElement, self.__http_client, f"/api/v1/synthetic/monitors", target_params=params, list_item="monitors")

    def get_full_monitor_configuration(self, monitor_id: str) -> SyntheticMonitor:
        """
        Get full monitor configuration for the specified monitor id (aka entity id).
        """
        return SyntheticMonitor(self.__http_client, raw_element=self.__http_client.make_request(f"/api/v1/synthetic/monitors/{monitor_id}").json())



from typing import List, Optional

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient


class GeoRegionsIpAddressMappingsService:
    ENDPOINT = "/api/config/v1/geographicRegions/ipAddressMappings"

    def __init__(self, http_client: HttpClient):
        self._http_client = http_client

    def list(self) -> "IPAddressMapping":
        """
        Lists all IP address mappings of the environment
        """
        response = self._http_client.make_request(self.ENDPOINT).json()
        return IPAddressMapping(self._http_client, None, response)

    def put(self, mappings: List["IPAddressMapping"]):
        """
        Updates the IP address mappings of the environment
        """
        data = [mapping.json() for mapping in mappings]
        body = {
            "ipAddressMappingRules": data
        }
        self._http_client.make_request(self.ENDPOINT, method="PUT", params=body)


class IPAddressMapping(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.rules: List["IPAddressMappingRule"] = [
            IPAddressMappingRule(raw_element=e) for e in raw_element.get("ipAddressMappingRules")
        ]


class IPAddressMappingRule(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.location = IPAddressMappingLocation(raw_element=raw_element.get("location"))
        self.ip_range = IPAddressRange(raw_element=raw_element.get("ipRange"))

    @staticmethod
    def create(location: "IPAddressMappingLocation", ip_range: "IPAddressRange"):
        """
        Creates a new IP address mapping rule
        """
        return IPAddressMappingRule(raw_element={
            "ipAddressMappingLocation": location.json(),
            "ipAddressRange": ip_range.json(),
        })


class IPAddressMappingLocation(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.country_code: str = raw_element.get("countryCode")
        self.city: Optional[str] = raw_element.get("city")
        self.region_code: Optional[str] = raw_element.get("regionCode")
        self.latitude: Optional[float] = raw_element.get("latitude")
        self.longitude: Optional[float] = raw_element.get("longitude")

    @staticmethod
    def create(country_code: str,
               city: Optional[str] = None,
               region_code: Optional[str] = None,
               latitude: Optional[float] = None,
               longitude: Optional[float] = None):
        return IPAddressMappingLocation(
            raw_element={
                "countryCode": country_code,
                "city": city,
                "regionCode": region_code,
                "latitude": latitude,
                "longitude": longitude,
            },
        )


class IPAddressRange(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.address: str = raw_element.get("address")
        self.subnet_mask: Optional[int] = raw_element.get("subnetMask")
        self.address_to: Optional[str] = raw_element.get("addressTo")

    @staticmethod
    def create(address: str, subnet_mask: Optional[int] = None, address_to: Optional[str] = None) -> "IPAddressRange":
        return IPAddressRange(raw_element={
            "address": address,
            "subnetMask": subnet_mask,
            "addressTo": address_to
        })



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from datetime import datetime
from enum import Enum
from typing import List, Optional, Union, Dict, Any

from requests import Response

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList
from dynatrace.utils import timestamp_to_string, int64_to_datetime


class MetricService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def query(
        self,
        metric_selector: str,
        resolution: str = None,
        time_from: Optional[Union[datetime, str]] = None,
        time_to: Optional[Union[datetime, str]] = None,
        entity_selector: Optional[str] = None,
        mz_selector: Optional[str] = None,
    ) -> PaginatedList["MetricSeriesCollection"]:

        params = {
            "metricSelector": metric_selector,
            "resolution": resolution,
            "from": timestamp_to_string(time_from),
            "to": timestamp_to_string(time_to),
            "entitySelector": entity_selector,
            "mzSelector": mz_selector
        }
        return PaginatedList(MetricSeriesCollection, self.__http_client, "/api/v2/metrics/query", params, list_item="result")

    def list(
        self,
        metric_selector: Optional[str] = None,
        text: Optional[str] = None,
        fields: Optional[str] = None,
        written_since: Optional[Union[str, datetime]] = None,
        metadata_selector: Optional[str] = None,
        page_size=100,
    ) -> PaginatedList["MetricDescriptor"]:
        params = {
            "pageSize": page_size,
            "metricSelector": metric_selector,
            "text": text,
            "fields": fields,
            "writtenSince": timestamp_to_string(written_since),
            "metadataSelector": metadata_selector,
        }
        return PaginatedList(MetricDescriptor, self.__http_client, "/api/v2/metrics", params, list_item="metrics")

    def get(self, metric_id: str) -> "MetricDescriptor":
        response = self.__http_client.make_request(f"/api/v2/metrics/{metric_id}").json()
        return MetricDescriptor(http_client=self.__http_client, raw_element=response)

    def delete(self, metric_id) -> Response:
        return self.__http_client.make_request(f"/api/v2/metrics/{metric_id}", method="DELETE")

    def ingest(self, lines: List[str]):
        lines = "\n".join(lines).encode("utf-8")
        return self.__http_client.make_request(
            f"/api/v2/metrics/ingest", method="POST", data=lines, headers={"Content-Type": "text/plain; charset=utf-8"}
        ).json()


class MetricSeries(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.timestamps: List[datetime] = [int64_to_datetime(timestamp) for timestamp in raw_element.get("timestamps", [])]
        self.dimensions: List[str] = raw_element.get("dimensions", [])
        self.values: List[float] = raw_element.get("values", [])
        self.dimension_map: Optional[Dict[str, Any]] = raw_element.get("dimensionMap", [])


class MetricSeriesCollection(DynatraceObject):
    def _create_from_raw_data(self, raw_element: dict):
        self.metric_id: str = raw_element.get("metricId")
        self.data: List[MetricSeries] = [MetricSeries(self._http_client, self._headers, metric_serie) for metric_serie in raw_element.get("data", [])]
        self.warnings: Optional[List[str]] = raw_element.get("warnings")


class MetricDefaultAggregation(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.parameter: float = raw_element.get("parameter")
        self.type: str = raw_element.get("type")


class MetricDimensionDefinition(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.index: int = raw_element.get("index")
        self.name: str = raw_element.get("name")
        self.key: str = raw_element.get("key")
        self.type: str = raw_element.get("type")


class AggregationType(Enum):
    AUTO = "auto"
    AVG = "avg"
    COUNT = "count"
    MAX = "max"
    MEDIAN = "median"
    MIN = "min"
    PERCENTILE = "percentile"
    SUM = "sum"
    VALUE = "value"


class Transformation(Enum):
    DEFAULT = "default"
    FILTER = "filter"
    FOLD = "fold"
    LAST = "last"
    LIMIT = "limit"
    MERGE = "merge"
    NAMES = "names"
    PARENTS = "parents"
    RATE = "rate"
    SORT = "sort"
    SPLITBY = "splitBy"
    TIMESHIFT = "timeshift"
    LASTREAL = "lastReal"
    SETUNIT = "setUnit"


class Unit(Enum):
    BIT = "Bit"
    BITPERHOUR = "BitPerHour"
    BITPERMINUTE = "BitPerMinute"
    BITPERSECOND = "BitPerSecond"
    BYTE = "Byte"
    BYTEPERHOUR = "BytePerHour"
    BYTEPERMINUTE = "BytePerMinute"
    BYTEPERSECOND = "BytePerSecond"
    CORES = "Cores"
    COUNT = "Count"
    DAY = "Day"
    DECIBELMILLIWATT = "DecibelMilliWatt"
    GIBIBYTE = "GibiByte"
    GIGA = "Giga"
    GIGABYTE = "GigaByte"
    HOUR = "Hour"
    KIBIBYTE = "KibiByte"
    KIBIBYTEPERHOUR = "KibiBytePerHour"
    KIBIBYTEPERMINUTE = "KibiBytePerMinute"
    KIBIBYTEPERSECOND = "KibiBytePerSecond"
    KILO = "Kilo"
    KILOBYTE = "KiloByte"
    KILOBYTEPERHOUR = "KiloBytePerHour"
    KILOBYTEPERMINUTE = "KiloBytePerMinute"
    KILOBYTEPERSECOND = "KiloBytePerSecond"
    MSU = "MSU"
    MEBIBYTE = "MebiByte"
    MEBIBYTEPERHOUR = "MebiBytePerHour"
    MEBIBYTEPERMINUTE = "MebiBytePerMinute"
    MEBIBYTEPERSECOND = "MebiBytePerSecond"
    MEGA = "Mega"
    MEGABYTE = "MegaByte"
    MEGABYTEPERHOUR = "MegaBytePerHour"
    MEGABYTEPERMINUTE = "MegaBytePerMinute"
    MEGABYTEPERSECOND = "MegaBytePerSecond"
    MICROSECOND = "MicroSecond"
    MILLICORES = "MilliCores"
    MILLISECOND = "MilliSecond"
    MILLISECONDPERMINUTE = "MilliSecondPerMinute"
    MINUTE = "Minute"
    MONTH = "Month"
    NANOSECOND = "NanoSecond"
    NANOSECONDPERMINUTE = "NanoSecondPerMinute"
    NOTAPPLICABLE = "NotApplicable"
    PERHOUR = "PerHour"
    PERMINUTE = "PerMinute"
    PERSECOND = "PerSecond"
    PERCENT = "Percent"
    PIXEL = "Pixel"
    PROMILLE = "Promille"
    RATIO = "Ratio"
    SECOND = "Second"
    STATE = "State"
    UNSPECIFIED = "Unspecified"
    WEEK = "Week"
    YEAR = "Year"


class MetricDescriptor(DynatraceObject):
    def _create_from_raw_data(self, raw_element):

        # required
        self.metric_id: str = raw_element.get("metricId")

        # optional
        self.aggregation_types: Optional[List[AggregationType]] = [AggregationType(element) for element in raw_element.get("aggregationTypes", [])]
        self.created: Optional[datetime] = int64_to_datetime(raw_element.get("created"))
        self.ddu_billable: Optional[bool] = raw_element.get("dduBillable")
        self.default_aggregation: Optional[MetricDefaultAggregation] = MetricDefaultAggregation(raw_element=raw_element.get("defaultAggregation"))
        self.description: Optional[str] = raw_element.get("description")
        self.dimension_definitions: Optional[List[MetricDimensionDefinition]] = [
            MetricDimensionDefinition(raw_element=element) for element in raw_element.get("dimensionDefinitions", [])
        ]
        self.display_name: Optional[str] = raw_element.get("displayName")
        self.entity_type: Optional[List[str]] = raw_element.get("entityType", [])
        self.impact_relevant: Optional[bool] = raw_element.get("impactRelevant")
        self.last_written: Optional[datetime] = int64_to_datetime(raw_element.get("lastWritten"))
        self.maximum_value: Optional[float] = raw_element.get("maximumValue")
        self.metric_value_type: Optional["MetricValueType"] = (
            MetricValueType(raw_element=raw_element.get("metricValueType")) if raw_element.get("metricValueType") else None
        )
        self.minimum_value: Optional[float] = raw_element.get("minimumValue")
        self.root_cause_relevant: Optional[bool] = raw_element.get("rootCauseRelevant")
        self.tags: Optional[List[str]] = raw_element.get("tags")
        self.transformations: Optional[List[Transformation]] = [Transformation(element) for element in raw_element.get("transformations", [])]
        self.unit: Optional[Unit] = Unit(raw_element.get("unit")) if raw_element.get("unit") else None
        self.warnings: Optional[List[str]] = raw_element.get("warnings")


class ValueType(Enum):
    ERROR = "error"
    SCORE = "score"
    UNKNOWN = "unknown"


class MetricValueType(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.type = ValueType(raw_element.get("type"))



from abc import ABC, abstractmethod
from base64 import b64encode
from typing import List, Optional

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList


class CredentialVaultService:
    _ENDPOINT = "/api/config/v1/credentials"

    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(self) -> PaginatedList["CredentialsResponseElement"]:
        """
        Lists all sets of credentials stored in your environment

        Required Api-Token scope:
            credentialVault.read (Read credential vault entries)
        """
        return PaginatedList(CredentialsResponseElement, self.__http_client, self._ENDPOINT, list_item="credentials")

    def post(self, credential: "Credentials"):
        response = self.__http_client.make_request(path=self._ENDPOINT,
                                                  params=credential.to_json(),
                                                  method="POST")

        return CredentialsResponseElement(raw_element=response.json())



class CredentialUsageHandler(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.type: str = raw_element.get("type", "")
        self.count: int = raw_element.get("count", 0)


class CredentialsResponseElement(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.name: str =  raw_element.get("name", "")
        self.id: str = raw_element.get("id", "")
        self.type: str = raw_element.get("type", "")
        self.description: str = raw_element.get("description", "")
        self.owner: str = raw_element.get("owner", "")
        self.owner_access_only = raw_element.get("ownerAccessOnly", False)
        self.scope: str = raw_element.get("scope", "")
        self.credential_usage_summary: List[CredentialUsageHandler] = [CredentialUsageHandler(raw_element=summary) for summary in raw_element.get("credentialUsageSummary", [])]


class Credentials(ABC):
    def __init__(self,
                 name: str,
                 scope: str,
                 description: Optional[str] = None,
                 owner_access_only: Optional[bool] = False,
                 credential_type: Optional[str] = None):
        self.name: str = name
        self.scope: str = scope
        self.description: Optional[str] = description
        self.owner_access_only: Optional[bool] = owner_access_only
        self.type: Optional[str] = credential_type

    @abstractmethod
    def to_json(self) -> dict:
        pass


class PublicCertificateCredentials(Credentials):
    def __init__(self,
                 name: str,
                 scope: str,
                 certificate: str,
                 password: str,
                 certificate_format: str,
                 description: Optional[str] = None,
                 owner_access_only: Optional[bool] = False,
                 credential_type: Optional[str] = None):
        super().__init__(name, scope, description, owner_access_only, credential_type)
        self.certificate: str = b64encode(certificate.encode()).decode()

        self.password: str = password
        self.certificate_format: str = certificate_format

    def to_json(self) -> dict:
        return {
            "name": self.name,
            "scope": self.scope,
            "description": self.description,
            "ownerAccessOnly": self.owner_access_only,
            "type": self.type,
            "certificate": self.certificate,
            "password": self.password,
            "certificateFormat": self.certificate_format
        }



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Optional, Dict, Any, Union
from enum import Enum
from requests import Response

from dynatrace.environment_v2.schemas import ConfigurationMetadata
from dynatrace.environment_v2.monitored_entities import EntityShortRepresentation
from dynatrace.pagination import PaginatedList
from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient


class AggregationType(Enum):
    AVG = "AVG"
    COUNT = "COUNT"
    MAX = "MAX"
    MEDIAN = "MEDIAN"
    MIN = "MIN"
    P90 = "P90"
    SUM = "SUM"
    VALUE = "VALUE"


class AlertCondition(Enum):
    ABOVE = "ABOVE"
    BELOW = "BELOW"


class DisabledReason(Enum):
    METRIC_DEFINITION_INCONSISTENCY = "METRIC_DEFINITION_INCONSISTENCY"
    NONE = "NONE"
    TOO_MANY_DIMS = "TOO_MANY_DIMS"
    TOPX_FORCIBLY_DEACTIVATED = "TOPX_FORCIBLY_DEACTIVATED"


class MetricEventAlertingScopeFilterType(Enum):
    ENTITY_ID = "ENTITY_ID"
    MANAGEMENT_ZONE = "MANAGEMENT_ZONE"
    TAG = "TAG"
    NAME = "NAME"
    CUSTOM_DEVICE_GROUP_NAME = "CUSTOM_DEVICE_GROUP_NAME"
    HOST_GROUP_NAME = "HOST_GROUP_NAME"
    HOST_NAME = "HOST_NAME"
    PROCESS_GROUP_ID = "PROCESS_GROUP_ID"
    PROCESS_GROUP_NAME = "PROCESS_GROUP_NAME"


class MetricEventDimensionsFilterType(Enum):
    ENTITY = "ENTITY"
    STRING = "STRING"


class MetricEventMonitoringStrategyType(Enum):
    STATIC_THRESHOLD = "STATIC_THRESHOLD"
    AUTO_ADAPTIVE_BASELINE = "AUTO_ADAPTIVE_BASELINE"


class Severity(Enum):
    AVAILABILITY = "AVAILABILITY"
    CUSTOM_ALERT = "CUSTOM_ALERT"
    ERROR = "ERROR"
    INFO = "INFO"
    PERFORMANCE = "PERFORMANCE"
    RESOURCE_CONTENTION = "RESOURCE_CONTENTION"


class Unit(Enum):
    BIT = "BIT"
    BIT_PER_HOUR = "BIT_PER_HOUR"
    BIT_PER_MINUTE = "BIT_PER_MINUTE"
    BIT_PER_SECOND = "BIT_PER_SECOND"
    BYTE = "BYTE"
    BYTE_PER_HOUR = "BYTE_PER_HOUR"
    BYTE_PER_MINUTE = "BYTE_PER_MINUTE"
    BYTE_PER_SECOND = "BYTE_PER_SECOND"
    CORES = "CORES"
    COUNT = "COUNT"
    DAY = "DAY"
    DECIBEL_MILLI_WATT = "DECIBEL_MILLI_WATT"
    GIBI_BYTE = "GIBI_BYTE"
    GIGA = "GIGA"
    GIGA_BYTE = "GIGA_BYTE"
    HOUR = "HOUR"
    KIBI_BYTE = "KIBI_BYTE"
    KIBI_BYTE_PER_HOUR = "KIBI_BYTE_PER_HOUR"
    KIBI_BYTE_PER_MINUTE = "KIBI_BYTE_PER_MINUTE"
    KIBI_BYTE_PER_SECOND = "KIBI_BYTE_PER_SECOND"
    KILO = "KILO"
    KILO_BYTE = "KILO_BYTE"
    KILO_BYTE_PER_HOUR = "KILO_BYTE_PER_HOUR"
    KILO_BYTE_PER_MINUTE = "KILO_BYTE_PER_MINUTE"
    KILO_BYTE_PER_SECOND = "KILO_BYTE_PER_SECOND"
    MEBI_BYTE = "MEBI_BYTE"
    MEBI_BYTE_PER_HOUR = "MEBI_BYTE_PER_HOUR"
    MEBI_BYTE_PER_MINUTE = "MEBI_BYTE_PER_MINUTE"
    MEBI_BYTE_PER_SECOND = "MEBI_BYTE_PER_SECOND"
    MEGA = "MEGA"
    MEGA_BYTE = "MEGA_BYTE"
    MEGA_BYTE_PER_HOUR = "MEGA_BYTE_PER_HOUR"
    MEGA_BYTE_PER_MINUTE = "MEGA_BYTE_PER_MINUTE"
    MEGA_BYTE_PER_SECOND = "MEGA_BYTE_PER_SECOND"
    MICRO_SECOND = "MICRO_SECOND"
    MILLI_CORES = "MILLI_CORES"
    MILLI_SECOND = "MILLI_SECOND"
    MILLI_SECOND_PER_MINUTE = "MILLI_SECOND_PER_MINUTE"
    MINUTE = "MINUTE"
    MONTH = "MONTH"
    MSU = "MSU"
    NANO_SECOND = "NANO_SECOND"
    NANO_SECOND_PER_MINUTE = "NANO_SECOND_PER_MINUTE"
    NOT_APPLICABLE = "NOT_APPLICABLE"
    PERCENT = "PERCENT"
    PER_HOUR = "PER_HOUR"
    PER_MINUTE = "PER_MINUTE"
    PER_SECOND = "PER_SECOND"
    PIXEL = "PIXEL"
    PROMILLE = "PROMILLE"
    RATIO = "RATIO"
    SECOND = "SECOND"
    STATE = "STATE"
    UNSPECIFIED = "UNSPECIFIED"
    WEEK = "WEEK"
    YEAR = "YEAR"


class WarningReason(Enum):
    NONE = "NONE"
    TOO_MANY_DIMS = "TOO_MANY_DIMS"


class MetricEventTextFilter(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.value: str = raw_element.get("value")
        self.operator: str = raw_element.get("operator")


class TagFilter(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.context: str = raw_element.get("context")
        self.key: str = raw_element.get("key")
        self.value: str = raw_element.get("value")


class MetricEventAlertingScope(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.filter_type: MetricEventAlertingScopeFilterType = MetricEventAlertingScopeFilterType(raw_element.get("filterType"))
        self.tag_filter: TagFilter = TagFilter(raw_element=raw_element.get("tagFilter"))
        self.name_filter: MetricEventTextFilter = MetricEventTextFilter(raw_element=raw_element.get("nameFilter"))
        self.entity_id: str = raw_element.get("entityId")
        self.management_zone_id: str = raw_element.get("mzId")
        self.process_group_id: str = raw_element.get("processGroupId")


class MetricEventDimension(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.filter_type: MetricEventDimensionsFilterType = MetricEventDimensionsFilterType(raw_element.get("filterType"))
        self.key: str = raw_element.get("key")
        self.name_filter: MetricEventTextFilter = MetricEventTextFilter(raw_element=raw_element.get("nameFilter"))
        self.text_filter: MetricEventTextFilter = MetricEventTextFilter(raw_element=raw_element.get("textFilter"))


class MetricEventMonitoringStrategy(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.type: MetricEventMonitoringStrategyType = MetricEventMonitoringStrategyType(raw_element.get("type"))


class MetricEventAutoAdaptiveBaselineMonitoringStrategy(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.type: MetricEventMonitoringStrategyType = MetricEventMonitoringStrategyType(raw_element.get("type"))
        self.samples: int = raw_element.get("samples")
        self.violating_samples: int = raw_element.get("violatingSamples")
        self.dealtering_samples: int = raw_element.get("dealertingSamples")
        self.alert_condition: AlertCondition = AlertCondition(raw_element.get("alertCondition"))
        self.alerting_on_missing_data: Optional[bool] = raw_element.get("alertingOnMissingData") if raw_element.get("alertingOnMissingData") else None
        self.number_of_signal_fluctuations: int = raw_element.get("numberOfSignalFluctuations")


class MetricEventStaticThresholdMonitoringStrategy(MetricEventMonitoringStrategy):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.type: MetricEventMonitoringStrategyType = MetricEventMonitoringStrategyType(raw_element.get("type"))
        self.samples: int = raw_element.get("samples")
        self.violating_samples: int = raw_element.get("violatingSamples")
        self.dealtering_samples: int = raw_element.get("dealertingSamples")
        self.alert_condition: AlertCondition = AlertCondition(raw_element.get("alertCondition"))
        self.alerting_on_missing_data: Optional[bool] = raw_element.get("alertingOnMissingData") if raw_element.get("alertingOnMissingData") else None
        self.threshold: int = raw_element.get("threshold")
        self.unit: Optional[Unit] = Unit(raw_element.get("unit")) if raw_element.get("unit") else None


class MetricEvent(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.metadata: ConfigurationMetadata = ConfigurationMetadata(raw_element=raw_element.get("metadata"))
        self.id: str = raw_element.get("id")
        self.metric_id: str = raw_element.get("metricId")
        self.name: str = raw_element.get("name")
        self.description: str = raw_element.get("description")
        self.aggregation_type: Optional[AggregationType] = AggregationType(raw_element.get("aggregationType")) if raw_element.get("aggregationType") else None
        self.severity: Severity = Severity(raw_element.get("severity"))
        self.enabled: bool = raw_element.get("enabled")
        self.disabled_reason: DisabledReason = DisabledReason(raw_element.get("disabledReason"))
        self.warning_reason: WarningReason = WarningReason(raw_element.get("warningReason"))
        self.alerting_scope: List[MetricEventAlertingScope] = [
            MetricEventAlertingScope(raw_element=raw_scope) for raw_scope in raw_element.get("alertingScope", [])
        ]
        self.metric_dimensions: List[MetricEventDimension] = [
            MetricEventDimension(raw_element=raw_dimension) for raw_dimension in raw_element.get("metricDimensions", [])
        ]
        self.monitoring_strategy: Union[
            MetricEventMonitoringStrategy, MetricEventStaticThresholdMonitoringStrategy, MetricEventAutoAdaptiveBaselineMonitoringStrategy
        ] = self._create_specific_monitoring_strategy(raw_element.get("monitoringStrategy"))
        self.primary_dimension_key: str = raw_element.get("primaryDimensionKey")

    def _create_specific_monitoring_strategy(self, raw_element: Dict[str, Any]):
        strategy_type = MetricEventMonitoringStrategyType(raw_element.get("type"))
        if strategy_type == MetricEventMonitoringStrategyType.AUTO_ADAPTIVE_BASELINE:
            return MetricEventAutoAdaptiveBaselineMonitoringStrategy(raw_element=raw_element)
        elif strategy_type == MetricEventMonitoringStrategyType.STATIC_THRESHOLD:
            return MetricEventStaticThresholdMonitoringStrategy(raw_element=raw_element)
        else:
            return MetricEventMonitoringStrategy(raw_element=raw_element)


class MetricEventShortRepresentation(EntityShortRepresentation):
    def delete(self) -> Response:
        """
        Deletes this metric event
        """
        return self._http_client.make_request(f"/api/config/v1/anomalyDetection/metricEvents/{self.id}", method="DELETE")

    def get_full_metric_event(self) -> MetricEvent:
        """
        Gets the full metric event for this stub
        """
        response = self._http_client.make_request(f"/api/config/v1/anomalyDetection/metricEvents/{self.id}").json()
        return MetricEvent(self._http_client, None, response)


class MetricEventService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(self) -> PaginatedList[MetricEventShortRepresentation]:
        """
        Lists all metric events in the environment. No configurable parameters.
        """
        return PaginatedList(MetricEventShortRepresentation, self.__http_client, f"/api/config/v1/anomalyDetection/metricEvents", list_item="values")



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from typing import Optional, Dict, Any, List, Union
from enum import Enum
from datetime import datetime

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient
from dynatrace.pagination import HeaderPaginatedList
from dynatrace.utils import datetime_to_int64

class RelativeTime(Enum):
    MIN = "min"
    FIVE_MINS = "5mins"
    TEN_MINS = "10mins"
    FIFTEEN_MINS = "15mins"
    THIRTY_MINS = "30mins"
    HOUR = "hour"
    TWO_HOURS = "2hours"
    SIX_HOURS = "6hours"
    DAY = "day"
    THREE_DAYS = "3days"


class OSArchitecture(Enum):
    ARM = "ARM"
    IA_SIXTY_FOUR = "IA64"
    PARISC = "PARISC"
    PPC = "PPC"
    PPCLE = "PPCLE"
    SYSTEM_THIRTY_NINTEY = "S390"
    SPARC = "SPARC"
    X_EIGHTY_SIX = "X86"
    ZOS = "ZOS"


class MonitoringMode(Enum):
    FULL_STACK = "FULL_STACK"
    INFRASTRUCTURE = "INFRASTRUCTURE"
    OFF = "OFF"
    NONE = None


class TagInfo(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.context: str = raw_element.get("context")
        self.key: str = raw_element.get("key")
        self.value: str = raw_element.get("value")


class AgentVersion(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.major: int = raw_element.get("major")
        self.minor: int = raw_element.get("minor")
        self.revision: int = raw_element.get("revision")
        self.timestamp: str = raw_element.get("timestamp")
        self.source_revision: str = raw_element.get("sourceRevision")


class HostGroup(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.me_id: str = raw_element.get("meId")
        self.name: str = raw_element.get("name")


class Host(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.entity_id: str = raw_element.get("entityId")
        self.display_name: str = raw_element.get("displayName")
        self.customized_name: str = raw_element.get("customizedName")
        self.discovered_name: str = raw_element.get("discoveredName")
        self.first_seen_timestamp: int = raw_element.get("firstSeenTimestamp")
        self.last_seen_timestamp: int = raw_element.get("lastSeenTimestamp")
        self.tags: List[TagInfo] = [TagInfo(raw_element=tag) for tag in raw_element.get("tags")]
        self.os_type: str = raw_element.get("osType")
        self.consumed_host_units: float = raw_element.get("consumedHostUnits")
        self.agent_version: AgentVersion = AgentVersion(raw_element=raw_element.get("agentVersion"))
        self.monitoring_mode: Union[MonitoringMode, None] = MonitoringMode(raw_element.get("monitoringMode"))
        self.network_zone_id: str = raw_element.get("networkZoneId")
        self.host_group: HostGroup = HostGroup(raw_element=raw_element.get("hostGroup"))
        self.os_architecture: OSArchitecture = OSArchitecture(raw_element.get("osArchitecture"))
        self.cpu_cores: int = raw_element.get("cpuCores")
        self.os_version: str = raw_element.get("osVersion")


class SmartScapeHostsService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(
        self,
        relative_time: Optional[Union[RelativeTime, str]] = RelativeTime.THREE_DAYS,
        start_timestamp: Optional[Union[datetime, str]] = None,
        end_timestamp: Optional[Union[datetime, str]] = None,
        page_size: int = 200,
        management_zone: Optional[int] = None,
        host_group_name: Optional[str] = None,
    ) -> HeaderPaginatedList[Host]:
        """
        List all monitored hosts

        :param management_zone: Filter hosts by a management zone ID
            Default value : None
        :param host_group_name: Filter hosts by a host group name
            Default value : None
        :param relative_time: Relative time ranger to check for (72 hours if not set)
            Default value : RelativeTime.THREE_DAYS
        :param start_timestamp: the start timestamp of the requested timeframe, in milliseconds (UTC)
        :param end_timestamp: the end timestamp of the requested timeframe, in milliseconds (UTC)
        """
        params = {
            "pageSize": page_size,
            "relativeTime": RelativeTime(relative_time).value if not start_timestamp else None,
            "startTimestamp": datetime_to_int64(start_timestamp),
            "endTimestamp": datetime_to_int64(end_timestamp),
            "managementZone": management_zone if management_zone else None,
            "hostGroupName": host_group_name if host_group_name else None,
        }
        return HeaderPaginatedList(Host, self.__http_client, f"/api/v1/entity/infrastructure/hosts", params)



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from enum import Enum
from typing import List, Optional, Union

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList
from dynatrace.environment_v2.schemas import VersionCompareType


class OsType(Enum):
    LINUX = "LINUX"
    WINDOWS = "WINDOWS"


class ActivegateType(Enum):
    ENVIRONMENT = "ENVIRONMENT"
    ENVIRONMENT_MULTI = "ENVIRONMENT_MULTI"


class UpdateStatus(Enum):
    INCOMPATIBLE = "INCOMPATIBLE"
    OUTDATED = "OUTDATED"
    SUPPRESSED = "SUPPRESSED"
    UNKNOWN = "UNKNOWN"
    UP2DATE = "UP2DATE"
    UPDATE_IN_PROGRESS = "UPDATE_IN_PROGRESS"
    UPDATE_PENDING = "UPDATE_PENDING"
    UPDATE_PROBLEM = "UPDATE_PROBLEM"


class AutoUpdate(Enum):
    ENABLED = "ENABLED"
    DISABLED = "DISABLED"


class Module(Enum):
    AWS = "AWS"
    AZURE = "AZURE"
    BEACON_FORWARDER = "BEACON_FORWARDER"
    CLOUD_FOUNDRY = "CLOUD_FOUNDRY"
    DB_INSIGHT = "DB_INSIGHT"
    EXTENSIONS_V1 = "EXTENSIONS_V1"
    EXTENSIONS_V2 = "EXTENSIONS_V2"
    KUBERNETES = "KUBERNETES"
    LOGS = "LOGS"
    MEMORY_DUMPS = "MEMORY_DUMPS"
    METRIC_API = "METRIC_API"
    ONE_AGENT_ROUTING = "ONE_AGENT_ROUTING"
    OTLP_INGEST = "OTLP_INGEST"
    REST_API = "REST_API"
    SYNTHETIC = "SYNTHETIC"


class ActiveGateService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(
        self,
        hostname: Optional[str] = None,
        os_type: Optional[Union[OsType, str]] = None,
        network_address: Optional[str] = None,
        activegate_type: Optional[Union[ActivegateType, str]] = None,
        network_zone: Optional[str] = None,
        update_status: Optional[Union[UpdateStatus, str]] = None,
        version_compare_type: Optional[Union[VersionCompareType, str]] = None,
        version: Optional[str] = None,
        auto_update: Optional[Union[AutoUpdate, str]] = None,
        group: Optional[str] = None,
        online: Optional[bool] = None,
        enabled_modules: Optional[List[Union[Module, str]]] = None,
        disabled_modules: Optional[List[Union[Module, str]]] = None,
    ) -> PaginatedList["ActiveGate"]:
        """
        Lists all available ActiveGates
        :return: A list of ActiveGates.
        """
        params = {
            "hostname": hostname,
            "osType": os_type,
            "networkAddress": network_address,
            "ActivegateType": activegate_type,
            "networkZone": network_zone,
            "updateStatus": update_status,
            "versionCompareType": version_compare_type,
            "version": version,
            "autoUpdate": auto_update,
            "group": group,
            "online": online,
            "enabledModule": enabled_modules,
            "disabledModule": disabled_modules,
        }
        return PaginatedList(ActiveGate, self.__http_client, "/api/v2/activeGates", params, list_item="activeGates")

    def get(self, activegate_id: str) -> "ActiveGate":
        return ActiveGate(raw_element=self.__http_client.make_request(f"/api/v2/activeGates/{activegate_id}").json())


class ActiveGate(DynatraceObject):
    def _create_from_raw_data(self, raw_element: dict):
        self.id: str = raw_element.get("id")
        self.network_addresses = raw_element.get("networkAddresses", [])
        self.os_type: str = raw_element.get("osType")
        self.auto_update_status: str = raw_element.get("autoUpdateStatus")
        self.offline_since: int = raw_element.get("offlineSince")
        self.version: str = raw_element.get("version")
        self.type: str = raw_element.get("type")
        self.hostname: str = raw_element.get("hostname")
        self.main_environment: str = raw_element.get("mainEnvironment")
        self.environments: str = raw_element.get("environments", [])
        self.network_zone: str = raw_element.get("networkZone")
        self.group: str = raw_element.get("group")
        self.modules: List["ActiveGateModule"] = [ActiveGateModule(raw_element=module) for module in raw_element.get("modules")]


class ActiveGateModule(DynatraceObject):
    def _create_from_raw_data(self, raw_element: dict):
        self.misconfigured: bool = raw_element.get("misconfigured")
        self.type: str = raw_element.get("type")
        self.attributes: bool = raw_element.get("attributes")
        self.enabled: bool = raw_element.get("enabled")
        self.version: str = raw_element.get("version")



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from enum import Enum
from typing import Optional, Dict, Any, List
from requests import Response

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.http_client import HttpClient
from dynatrace.pagination import PaginatedList

from dynatrace.environment_v2.custom_tags import METag
from dynatrace.environment_v2.monitored_entities import EntityShortRepresentation


class TagCombination(Enum):
    AND = "AND"
    OR = "OR"
    NONE = None

    def __str__(self) -> str:
        return self.value


class MonitoredEntityFilter(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.type: str = raw_element.get("type")
        self.mz_id: str = raw_element.get("mzId")
        self.tags: Optional[List[METag]] = [METag(raw_element=tag) for tag in raw_element.get("tags", [])]
        self.tag_combination: Optional[TagCombination] = TagCombination(raw_element.get("tagCombination"))

    def to_json(self) -> Dict[str, Any]:
        return {"type": self.type, "mzId": self.mz_id, "tags": [t.to_json() for t in self.tags], "tagCombination": str(self.tag_combination)}


class Scope(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.entities: List[str] = raw_element.get("entities")
        self.matches: Optional[List[MonitoredEntityFilter]] = [MonitoredEntityFilter(raw_element=m) for m in raw_element.get("matches")]

    def to_json(self) -> Dict[str, Any]:
        return {"entities": self.entities, "matches": [m.to_json() for m in self.matches]}


class Recurrence(DynatraceObject):
    @staticmethod
    def create(start_time: str, duration: int, day_of_week: Optional[str] = None, day_of_month: Optional[int] = None):
        raw_element = {"dayOfWeek": day_of_week.upper(), "dayOfMonth": day_of_month, "startTime": start_time, "durationMinutes": duration}
        return Recurrence(raw_element=raw_element)

    def _create_from_raw_data(self, raw_element):
        self.day_of_week: str = raw_element.get("dayOfWeek")
        self.day_of_month: str = raw_element.get("dayOfMonth")
        self.start_time: Optional[str] = raw_element.get("startTime")
        self.duration: Optional[int] = raw_element.get("durationMinutes")


class Schedule(DynatraceObject):
    @staticmethod
    def create(recurrence_type: str, start: str, end: str, zone_id: str, recurrence: Optional[Recurrence] = None):
        raw_element = {"recurrenceType": recurrence_type, "recurrence": recurrence, "start": start, "end": end, "zoneId": zone_id}
        return Schedule(raw_element=raw_element)

    @property
    def schedule_snippet(self) -> Dict[str, Any]:
        representation = {}
        if self.recurrence_type == "DAILY":
            representation = {
                "recurrenceType": self.recurrence_type,
                "recurrence": {"startTime": self.recurrence.start_time, "durationMinutes": self.recurrence.duration},
                "start": self.start_time,
                "end": self.end_time,
                "zoneId": self.zone_id,
            }
        if self.recurrence_type == "WEEKLY":
            representation = {
                "recurrenceType": self.recurrence_type,
                "recurrence": {"dayOfWeek": self.recurrence.day_of_week, "startTime": self.recurrence.start_time, "durationMinutes": self.recurrence.duration},
                "start": self.start_time,
                "end": self.end_time,
                "zoneId": self.zone_id,
            }
        if self.recurrence_type == "MONTHLY":
            representation = {
                "recurrenceType": self.recurrence_type,
                "recurrence": {
                    "dayOfMonth": self.recurrence.day_of_month,
                    "startTime": self.recurrence.start_time,
                    "durationMinutes": self.recurrence.duration,
                },
                "start": self.start_time,
                "end": self.end_time,
                "zoneId": self.zone_id,
            }
        if self.recurrence_type == "ONCE":
            representation = {"recurrenceType": self.recurrence_type, "start": self.start_time, "end": self.end_time, "zoneId": self.zone_id}
        return representation

    def _create_from_raw_data(self, raw_element):
        self.recurrence_type: str = raw_element.get("recurrenceType")
        self.recurrence: Recurrence = Recurrence(raw_element=raw_element.get("recurrence"))
        self.start_time: str = raw_element.get("start")
        self.end_time: str = raw_element.get("end")
        self.zone_id: str = raw_element.get("zoneId")


class MaintenanceWindow(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.description: str = raw_element.get("description")
        self.type: str = raw_element.get("type")
        self.suppression: str = raw_element.get("suppression")
        self.suppress_synthetic_monitors_execution: bool = raw_element.get("suppressSyntheticMonitorsExecution")
        self.scope: Scope = Scope(raw_element=raw_element.get("scope"))
        self.schedule: Schedule = Schedule(raw_element=raw_element.get("schedule"))

    def post(self) -> EntityShortRepresentation:
        """Creates the Maintenance Window configuration in Dynatrace (POST).

        :param maintenance_window: the Maintenance Window configuration details

        :returns EntityShortRepresentation: basic details of the created Maintenance Window

        :throws ValueError: if operation cannot be executed due to missing HTTP Client
        """
        if not self._http_client:
            raise ValueError("Object does not have an HTTP Client. Use maintenance_window.post() instead.")
        response = self._http_client.make_request(path=MaintenanceWindowService.ENDPOINT, params=self.to_json(), method="POST")
        self.id = response.json().get("id")

        return EntityShortRepresentation(raw_element=response.json())

    def to_json(self) -> Dict[str, Any]:
        """Get a JSON (dict) representation of this config."""
        mw: Dict[str, Any] = {
            "id": self.id,
            "name": self.name,
            "description": self.description,
            "type": self.type,
            "suppression": self.suppression,
            "suppressSyntheticMonitorsExecution": self.suppress_synthetic_monitors_execution,
            "schedule": self.schedule.schedule_snippet,
            "scope": self.scope.to_json(),
        }
        return mw


class MaintenanceWindowService:
    ENDPOINT = "/api/config/v1/maintenanceWindows"

    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def list(self) -> PaginatedList["MaintenanceWindowStub"]:
        """
        Lists all maintenance windows in the environment. No configurable parameters.
        """
        return PaginatedList(MaintenanceWindowStub, self.__http_client, self.ENDPOINT, list_item="values")

    def get(self, mw_id: str) -> MaintenanceWindow:
        """Gets the full details of the Maintenance Window referenced by ID.

        :param mw_id: ID of the alerting profile

        :returns AlertingProfile: alerting profile details
        """
        response = self.__http_client.make_request(f"{self.ENDPOINT}/{mw_id}")
        return MaintenanceWindow(http_client=self.__http_client, raw_element=response.json())

    def post(self, mw: MaintenanceWindow) -> EntityShortRepresentation:
        """Creates the Maintenance Window configuration in Dynatrace (POST).

        :param maintenace_window: the Maintenance Window configuration details

        :returns EntityShortRepresentation: basic details of the created Maintenance Window
        """
        if not mw._http_client:
            mw._http_client = self.__http_client
        return mw.post()

    def create_schedule(
        self,
        recurrence_type: str,
        start: str,
        end: str,
        zone_id: str,
        recurrence_start_time: Optional[str] = None,
        recurrence_duration: Optional[int] = None,
        recurrence_day_of_week: Optional[str] = None,
        recurrence_day_of_month: Optional[int] = None,
    ) -> "Schedule":
        """
        Create a schedule to be used when creating a maintenance window.
        """
        recurrence = (
            Recurrence.create(recurrence_start_time, recurrence_duration, recurrence_day_of_week, recurrence_day_of_month)
            if recurrence_type != "ONCE"
            else None
        )
        return Schedule.create(recurrence_type, start, end, zone_id, recurrence)

    def create(
        self,
        name: str,
        description: str,
        window_type: str,
        suppression: str,
        schedule: Schedule,
        maintenance_window_id: Optional[str] = None,
        suppress_synthetic: Optional[bool] = False,
        scope: Optional[Scope] = None,
        recurrence: Optional[Recurrence] = None,
    ) -> "MaintenanceWindowCreated":
        # TODO - scope and recurrence are not used here but they should
        """
        Create a maintenance window with the specified parameters.
        You must first use create_schedule() to create a schedule with optional recurrence.
        """
        body = {
            "id": maintenance_window_id,
            "name": name,
            "description": description,
            "type": window_type,
            "suppression": suppression,
            "suppressSyntheticMonitorsExecution": suppress_synthetic,
            "schedule": schedule.schedule_snippet,
            "scope": {"entities": scope.entities, "matches": [s.to_json() for s in scope.matches]},
        }

        response = self.__http_client.make_request(self.ENDPOINT, method="POST", params=body).json()
        return MaintenanceWindowCreated(raw_element=response)

    def delete(self, zone_id: str) -> Response:
        """
        Delete the maintenance window with the specified id
        """
        return self.__http_client.make_request(f"{self.ENDPOINT}/{zone_id}", method="DELETE")


class MaintenanceWindowCreated(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.description: str = raw_element.get("description")


class MaintenanceWindowStub(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.id: str = raw_element.get("id")
        self.name: str = raw_element.get("name")
        self.description: str = raw_element.get("description")

    def get_full_maintenance_window(self) -> MaintenanceWindow:
        """
        Gets the full maintenance window for this stub
        """
        response = self._http_client.make_request(f"{MaintenanceWindowService.ENDPOINT}/{self.id}").json()
        return MaintenanceWindow(self._http_client, None, response)



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict, Any

from requests import Response

from dynatrace.dynatrace_object import DynatraceObject
from dynatrace.environment_v2.schemas import ConfigurationMetadata
from dynatrace.http_client import HttpClient


class ActiveGateAutoUpdateConfigurationService:
    def __init__(self, http_client: HttpClient):
        self.__http_client = http_client

    def get_global(self) -> "ActiveGateGlobalAutoUpdateConfig":
        return ActiveGateGlobalAutoUpdateConfig(raw_element=self.__http_client.make_request("/api/v2/activeGates/autoUpdate").json())

    def put_global(self, setting: str) -> Response:
        """
        Puts the global auto-update configuration of environment ActiveGates

        :param setting: The state of auto-updates for all ActiveGates connected to the environment or Managed cluster. ENABLED or DISABLED
        """
        return self.__http_client.make_request("/api/v2/activeGates/autoUpdate", method="PUT", params={"globalSetting": setting})

    def validate_global(self, setting: str) -> Response:
        return self.__http_client.make_request("/api/v2/activeGates/autoUpdate/validator", method="POST", params={"globalSetting": setting})

    def get(self, activegate_id: str) -> "ActiveGateAutoUpdateConfig":
        return ActiveGateAutoUpdateConfig(raw_element=self.__http_client.make_request(f"/api/v2/activeGates/{activegate_id}/autoUpdate").json())

    def put(self, activegate_id: str, setting: str) -> Response:
        return self.__http_client.make_request(f"/api/v2/activeGates/{activegate_id}/autoUpdate", method="PUT", params={"setting": setting})

    def validate(self, activegate_id: str, setting: str) -> Response:
        return self.__http_client.make_request(f"/api/v2/activeGates/{activegate_id}/autoUpdate/validator", method="POST", params={"setting": setting})


class ActiveGateGlobalAutoUpdateConfig(DynatraceObject):
    def _create_from_raw_data(self, raw_element):
        self.global_setting: str = raw_element.get("globalSetting")
        self.metadata: ConfigurationMetadata = ConfigurationMetadata(raw_element=raw_element.get("metadata"))


class ActiveGateAutoUpdateConfig(DynatraceObject):
    def _create_from_raw_data(self, raw_element: Dict[str, Any]):
        self.setting: str = raw_element.get("setting")
        self.effective_setting: str = raw_element.get("effectiveSetting")



"""
Copyright 2021 Dynatrace LLC

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import logging
from typing import Dict, Optional, Any
import time

import requests
import urllib3
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


TOO_MANY_REQUESTS_WAIT = "wait"


class DynatraceRetry(Retry):
    def get_backoff_time(self):
        return self.backoff_factor


class HttpClient:
    def __init__(
        self,
        base_url: str,
        token: str,
        log: logging.Logger = None,
        proxies: Dict = None,
        too_many_requests_strategy=None,
        retries: int = 0,
        retry_delay_ms: int = 0,
        mc_jsession_id: Optional[str] = None,
        mc_b925d32c: Optional[str] = None,
        mc_sso_csrf_cookie: Optional[str] = None,
        print_bodies: bool = False,
        timeout: Optional[int] = None
    ):
        while base_url.endswith("/"):
            base_url = base_url[:-1]
        self.base_url = base_url

        if proxies is None:
            proxies = {}
        self.proxies = proxies

        self.auth_header = {"Authorization": f"Api-Token {token}"}
        self.print_bodies = print_bodies
        self.log = log
        if self.log is None:
            self.log = logging.getLogger(__name__)
            self.log.setLevel(logging.WARNING)
            st = logging.StreamHandler()
            fmt = logging.Formatter("%(asctime)s - %(levelname)s - %(name)s - %(thread)d - %(filename)s:%(lineno)d - %(message)s")
            st.setFormatter(fmt)
            self.log.addHandler(st)

        self.too_many_requests_strategy = too_many_requests_strategy
        self.timeout = timeout
        retry_delay_s = retry_delay_ms / 1000

        try:
            self.retries = DynatraceRetry(
                total=retries,
                backoff_factor=retry_delay_s,
                status_forcelist=[400, 401, 403, 404, 413, 429, 500, 502, 503, 504],
                allowed_methods=["TRACE", "PUT", "DELETE", "OPTIONS", "HEAD", "GET", "POST"],
                raise_on_status=False,
            )
        except TypeError:  # Older version of urllib3?
            self.retries = DynatraceRetry(
                total=retries,
                backoff_factor=retry_delay_s,
                status_forcelist=[400, 401, 403, 404, 413, 429, 500, 502, 503, 504],
                method_whitelist=["TRACE", "PUT", "DELETE", "OPTIONS", "HEAD", "GET", "POST"],
                raise_on_status=False,
            )

        # This is for internal dynatrace usage
        self.mc_jsession_id = mc_jsession_id
        self.mc_b925d32c = mc_b925d32c
        self.mc_sso_csrf_cookie = mc_sso_csrf_cookie

    def make_request(
        self, path: str, params: Optional[Any] = None, headers: Optional[Dict] = None, method="GET", data=None, files=None, query_params=None
    ) -> requests.Response:
        url = f"{self.base_url}{path}"

        body = None
        if method in ["POST", "PUT"]:
            body = params
            params = query_params

        if headers is None:
            headers = {}
        if files is None and "content-type" not in [key.lower() for key in headers.keys()]:
            headers.update({"content-type": "application/json"})
        headers.update(self.auth_header)

        cookies = None
        if self.mc_b925d32c and self.mc_sso_csrf_cookie and self.mc_jsession_id:
            headers.update({"Cookie": f"JSESSIONID={self.mc_jsession_id}; ssoCSRFCookie={self.mc_sso_csrf_cookie}; b925d32c={self.mc_b925d32c}"})
            cookies = {"JSESSIONID": self.mc_jsession_id, "ssoCSRFCookie": self.mc_sso_csrf_cookie, "b925d32c": self.mc_b925d32c}

        s = requests.Session()
        s.mount("https://", HTTPAdapter(max_retries=self.retries))

        self.log.debug(f"Making {method} request to '{url}' with params {params} and body: {body}")
        if self.print_bodies:
            print(method, url)
            if body:
                print(json.dumps(body, indent=2))
        r = s.request(method, url, headers=headers, params=params, json=body, verify=False, proxies=self.proxies, data=data, cookies=cookies, files=files, timeout=self.timeout)
        self.log.debug(f"Received response '{r}'")

        while r.status_code == 429 and self.too_many_requests_strategy == TOO_MANY_REQUESTS_WAIT:
            sleep_amount = int(r.headers.get("retry-after", 5))
            self.log.warning(f"Sleeping for {sleep_amount}s because we have received an HTTP 429")
            time.sleep(sleep_amount)
            r = requests.request(method, url, headers=headers, params=params, json=body, verify=False, proxies=self.proxies, timeout=self.timeout)

        if r.status_code >= 400:
            raise Exception(f"Error making request to {url}: {r}. Response: {r.text}")

        return r
